# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject

from .file import EartagFile


class EartagFileMutagenCommon(EartagFile):
    """Base class for Mutagen-based backends."""

    __gtype_name__ = "EartagFileMutagenCommon"

    def __init__(self, path):
        super().__init__(path)
        self.mg_file = None

    def save(self):
        """Saves the changes to the file."""
        self.mg_file.save()
        self.setup_original_values()
        self.mark_as_unmodified()

    def on_remove(self, *args):
        super().on_remove()
        self.mg_file = None

    def delete_all_raw(self):
        self.delete_all()
        if not self.mg_file.tags:
            return
        for tag in dict(self.mg_file.tags).keys():
            del self.mg_file.tags[tag]

    # Main properties

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def length(self):
        return int(self.mg_file.info.length)

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def bitrate(self):
        # in bps, needs conversion
        try:
            return int(round(self.mg_file.info.bitrate / 1000, 0))
        except AttributeError:
            # For some files, Mutagen can't tell the bitrate
            return -1

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def channels(self):
        try:
            return self.mg_file.info.channels
        except AttributeError:
            # For some files, Mutagen can't tell the channel count
            return 0



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from itertools import groupby
import re
import unicodedata
import os


def all_equal(iterable):
    """
    Check if all elements in a list are equal. Source:
    https://stackoverflow.com/questions/3844801/check-if-all-elements-in-a-list-are-identical
    """
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def find_in_model(model, item):
    """
    Gets the position of an item in the model, or -1 if not found.
    Replacement for .find function in models that don't have it.
    """
    i = 0
    while True:
        found = model.get_item(i)
        if not found:
            break
        if found == item:
            return i
        i += 1
    return -1


def is_float(value):
    """Checks if the given value is a valid float."""
    try:
        float(value)
    except ValueError:
        return False
    return True


def get_readable_length(length):
    """Returns human-readable version of the length, given in seconds."""
    length_min, length_sec = divmod(int(length), 60)
    length_hour, length_min = divmod(length_min, 60)

    if length_hour:
        length_readable = "{h}∶{m}∶{s}".format(
            h=str(length_hour).rjust(2, "0"),
            m=str(length_min).rjust(2, "0"),
            s=str(length_sec).rjust(2, "0"),
        )
    else:
        length_readable = "{m}∶{s}".format(
            m=str(length_min).rjust(2, "0"), s=str(length_sec).rjust(2, "0")
        )

    return length_readable


def title_case_preserve_uppercase(text: str):
    return " ".join([x.isupper() and x or x.capitalize() for x in text.split(" ")])


def simplify_string(text: str):
    """
    Returns a "simplified string" that throws away non-alphanumeric
    characters for more accurate searches and comparisons.
    """
    # Step 1: Normalize Unicode characters
    instr = unicodedata.normalize("NFKC", text)
    # Step 2: Only leave lowercase alphanumeric letters
    instr = "".join(
        [letter for letter in instr.lower() if letter.isalnum() or letter == " "]
    ).strip()
    # Step 3: Remove repeating spaces
    instr = re.sub(" +", " ", instr)

    return instr


def simplify_compare(string1: str, string2: str):
    """
    Compares simplified representations of two strings and returns
    whether they're equal.
    """
    return simplify_string(string1) == simplify_string(string2)


def reg_and_simple_cmp(string1: str, string2: str):
    return string1 == string2 or simplify_compare(string1, string2)


def inspect_prettyprint(stack):
    """
    Convenience function that pretty-prints the results of inspect.stack().

    This is not used anywhere in the program; it's only here for debugging
    purposes, since inspect is a pretty useful tool for finding optimization
    issues, and ends up being used quite frequently.
    """
    print("--- Inspect trace: ---")

    for frame in stack:
        print(f"\033[1m{frame.filename}:{frame.lineno}\033[0m")
        print("".join(frame.code_context))  # already includes newlines

    print("--- End trace ---")


# https://stackoverflow.com/questions/1976007/what-characters-are-forbidden-in-windows-and-linux-directory-names#31976060
FILENAME_BANNED_CHARS = ("/", "\\", ":", "*", "?", '"', "<", ">", "|", "\t")
BANNED_FILENAMES = (
    "CON",
    "PRN",
    "AUX",
    "NUL",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "COM5",
    "COM6",
    "COM7",
    "COM8",
    "COM9",
    "LPT1",
    "LPT2",
    "LPT3",
    "LPT4",
    "LPT5",
    "LPT6",
    "LPT7",
    "LPT8",
    "LPT9",
)


def cleanup_filename(filename: str, allow_path: bool = False, full_path: bool = False) -> str:
    """
    Takes a filename or path and returns a cleaned up version (that is,
    one where invalid characters or tricks like ../ are ignored).
    """
    if full_path:
        allow_path = True

    banned_chars = list(FILENAME_BANNED_CHARS)
    if allow_path and os.path.sep in banned_chars:
        banned_chars.remove(os.path.sep)

    for char in banned_chars:
        filename = filename.replace(char, "_")

    if full_path:
        filename_clean = os.path.sep
    else:
        if filename.startswith(os.path.sep):
            filename = filename[1:]
        filename_clean = ""

    filename_split = filename.split(os.path.sep)
    for i in range(len(filename_split)):
        # Strip out leading/trailing spaces
        f = filename_split[i].strip()

        if f == "":
            continue

        # Fix too long filenames; 255 is a common limit, we leave some
        # space for safety
        if len(f) > 249:
            if i == (len(filename_split) - 1):
                # If we're dealing with a filename, keep the extension
                f, ext = os.path.splitext(f)
                f = f[(-249 + len(ext) - 1) :] + ext
            else:
                f = f[-249:]

        # Misc. cleanup steps
        if f in BANNED_FILENAMES:
            filename_clean += f"_{f}"
        elif f == "..":
            filename_clean += "__"
        elif f.endswith("."):
            filename_clean += f[:-1] + "_"
        else:
            filename_clean += f
        filename_clean += os.path.sep
    filename_clean = filename_clean[:-1]

    return filename_clean


def filename_valid(filename: str, allow_path: bool = False, full_path: bool = False) -> bool:
    """Returns whether or not a filename is valid."""
    if filename != os.path.normpath(filename):
        return False

    banned_chars = list(FILENAME_BANNED_CHARS)
    if allow_path and os.path.sep in banned_chars:
        banned_chars.remove(os.path.sep)

    if not full_path and filename.startswith(os.path.sep):
        return False

    for char in banned_chars:
        if char in filename:
            return False

    for f in filename.split(os.path.sep):
        if f in BANNED_FILENAMES + ("..",) or f.endswith(".") or f != f.strip() or len(f) > 255:
            return False

    return True


def file_is_sandboxed(path: str) -> bool:
    """
    Takes a path and returns whether or not the file is sandboxed.

    This is a best-guess based on the path, and has only really been tested
    to work under Flatpak.
    """
    return bool(re.match("^/run/user/[0-9].*/doc", path))


def natural_compare(a: str, b: str) -> int:
    """
    Compares two strings using the "natural/human sorting" algorithm.

    This returns an integer, satisfying similar constraints to GLib's collate
    functions: 0 if both strings are the same, 1 if a > b, and -1 if a < b.
    """
    # Adapted from https://stackoverflow.com/a/8940266
    convert = lambda text: int(text) if text.isdigit() else text.lower()  # noqa: E731
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]  # noqa: E731
    sort = sorted([a, b], key=alphanum_key)

    if sort[0] == sort[1]:
        return 0
    if sort[0] is a:
        return -1
    return 1


def safe_int(value: int | str | None) -> int | None:
    """Wrapper around int() that handles invalid values gracefully."""
    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        value = value.strip()
    elif not value:
        return 0

    try:
        return int(value)
    except ValueError:
        return 0


def safe_float(value: int | float | str | None) -> float | None:
    """Wrapper around int() that handles invalid values gracefully."""
    if isinstance(value, float):
        return value
    elif isinstance(value, str):
        value = value.strip()
    elif not value:
        return 0.0

    try:
        return float(value)
    except ValueError:
        return 0.0


def iter_selection_model(model, position: int = -1, n_items: int = -1):
    """Iterate over the selected items in a selection model."""
    if (position >= 0 and n_items < 0) or (n_items >= 0 and position < 0):
        raise ValueError("Both position and n_items must be set")

    if position >= 0 and n_items >= 0:
        selection = model.get_selection_range(position, n_items)
    else:
        selection = model.get_selection()
    for i in range(selection.get_size()):
        pos = selection.get_nth(i)
        yield model.get_item(pos)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from . import APP_GRESOURCE_PATH
from .backends.file import EartagFile, VALID_TAGS, EXTRA_TAGS
from .config import config
from .utils.asynctask import EartagAsyncTask
from .utils.extracttags import extract_tags_from_filename
from .utils.tagsyntaxhighlight import (
    EartagPlaceholderSyntaxHighlighter,
    attr_foreground_new,
    THEMES,
)
from .utils.previewselector import EartagPreviewSelectorButton  # noqa: F401
from .utils.tagselector import EartagTagSelectorButton  # noqa: F401
from .utils.misc import filename_valid

from gi.repository import Adw, Gtk, Gio, GObject, Pango
import re
import os.path


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/extract.ui")
class EartagExtractTagsDialog(Adw.Dialog):
    """Dialog for extracting selected files' tags from their filename."""

    __gtype_name__ = "EartagExtractTagsDialog"

    pattern_entry = Gtk.Template.Child()
    preview_entry = Gtk.Template.Child()
    preview_selector_button = Gtk.Template.Child()
    apply_button = Gtk.Template.Child()
    toggle_strip_common_suffixes = Gtk.Template.Child()

    rename_progress = Gtk.Template.Child()
    content_clamp = Gtk.Template.Child()

    tag_selector = Gtk.Template.Child()

    theme = GObject.Property(type=str, default="light")

    strip_common_suffixes = GObject.Property(type=bool, default=True)
    validation_passed = GObject.Property(type=bool, default=True)

    preview_selector_button = Gtk.Template.Child()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self._connections = []
        self.custom_syntax_highlight = EartagPlaceholderSyntaxHighlighter(
            self.pattern_entry, "entry"
        )
        self.custom_syntax_highlight.bind_property(
            "error", self, "validation-passed", GObject.BindingFlags.SYNC_CREATE
        )
        self.custom_syntax_highlight.connect("notify::error", self.check_for_errors)
        self.connect("notify::validation-passed", self.on_syntax_highlight_error)

        self._entry_conn = self.pattern_entry.connect("changed", self.update_preview)

        self.custom_syntax_highlight.bind_property(
            "theme",
            self,
            "theme",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self._connections.append(self.connect("notify::theme", self.update_preview))

        self.files = parent.file_manager.selected_files_list.copy()

        self.preview_selector_button.set_files(self.files)
        self.preview_selector_button.set_formatting_function(self.generate_preview_attrs)
        self._preview_update_conn = self.preview_selector_button.connect(
            "notify::selected-index", self.update_preview
        )

        config.bind(
            "extract-strip-common-suffixes",
            self,
            "strip-common-suffixes",
            Gio.SettingsBindFlags.DEFAULT,
        )
        self.bind_property(
            "strip-common-suffixes",
            self.toggle_strip_common_suffixes,
            "active",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self._connections.append(self.connect("notify::strip-common-suffixes", self.update_preview))

        config.bind("extract-pattern", self.pattern_entry, "text", Gio.SettingsBindFlags.DEFAULT)

        self.apply_task = EartagAsyncTask(self.apply_func)
        self.apply_task.bind_property("progress", self.rename_progress, "fraction")
        self.apply_task.connect("task-done", self.on_apply_done)

        # Custom filter for tag selector to filter out already present tags
        self.tag_filter = Gtk.CustomFilter.new(self.tag_filter_func, self.tag_selector.tag_model)
        self.tag_selector.set_filter(self.tag_filter)

        self.pattern_entry.connect("changed", self.tag_selector.refresh_tag_filter)
        self.pattern_entry.connect("changed", self.check_for_errors)

        self.check_for_errors()
        self.update_preview()
        self.tag_selector.refresh_tag_filter()

    @property
    def present_tags(self) -> list:
        tags = re.findall(r"{(.*?)}", self.pattern_entry.get_text())
        tags = [x for x in set(tags) if x in VALID_TAGS]
        return tags

    def tag_filter_func(self, _tag_name, *args):
        """Filter function for the tag dropdown."""
        if not self.tag_selector.tag_filter_func(_tag_name):
            return False
        tag = self.tag_selector.tag_names_swapped[_tag_name.get_string()]
        if tag in ("length", "bitrate"):
            return False
        if tag in self.present_tags:
            return False
        return True

    @Gtk.Template.Callback()
    def add_tag_from_selector(self, selector, tag, *args):
        self.pattern_entry.insert_text("{" + tag + "}", self.pattern_entry.props.cursor_position)

    def on_syntax_highlight_error(self, *args):
        if not self.props.validation_passed:
            self.pattern_entry.add_css_class("error")
            self.apply_button.set_sensitive(False)  # todo: only do this when custom is selected
        else:
            self.pattern_entry.remove_css_class("error")
            self.apply_button.set_sensitive(True)  # todo: only do this when custom is selected

    def get_extracted(self, filename: str, positions: bool = False) -> dict:
        """
        Applies the pattern from pattern entry to the given filename
        and returns a guess (see extract_tags_from_filename function docs
        for more information).
        """
        extracted = extract_tags_from_filename(
            filename,
            self.pattern_entry.get_text(),
            positions=positions,
            strip_common_suffixes=self.props.strip_common_suffixes,
        )

        return extracted

    def generate_preview_attrs(self, file: EartagFile):
        filename = os.path.basename(file.props.path)
        extracted = self.get_extracted(filename, positions=True)

        preview_attrs = Pango.AttrList()
        self.preview_entry.set_text(filename)

        present_tags = []
        n = 0
        for tag, tag_data in extracted.items():
            if tag in present_tags:
                continue
            present_tags.append(tag)
            span = tag_data[1]
            color = THEMES[self.props.theme]["placeholder_colors"][
                n % len(THEMES[self.props.theme]["placeholder_colors"])
            ]
            color_attr = attr_foreground_new(color, span[0], span[1])
            preview_attrs.insert(color_attr)
            n += 1

        self.preview_entry.notify("attributes")
        self.preview_entry.set_attributes(preview_attrs)

        return (filename, preview_attrs)

    def update_preview(self, *args):
        """Updates the text and syntax highlighting on the preview entry."""
        self.preview_selector_button.emit("formatting-changed")
        preview_text, preview_attrs = self.generate_preview_attrs(
            self.files[self.preview_selector_button.props.selected_index]
        )
        self.preview_entry.set_text(preview_text)
        self.preview_entry.set_attributes(preview_attrs)

    @Gtk.Template.Callback()
    def do_apply(self, *args):
        if self.apply_button.props.sensitive is False:
            return
        if len(self.files) < 1:
            self.close()
        self.content_clamp.set_sensitive(False)
        self.apply_button.set_sensitive(False)
        self.set_can_close(False)
        self.apply_task.run()

    async def apply_func(self):
        self.extracted = 0

        progress_step = 1 / len(self.files)

        for file in self.files:
            filename = os.path.basename(file.path)
            extract = self.get_extracted(filename, positions=False)
            if not extract:
                continue

            for tag, value in extract.items():
                if tag not in EXTRA_TAGS or (
                    tag in EXTRA_TAGS and tag in file.supported_extra_tags
                ):
                    if tag in file.int_properties:
                        try:
                            file.set_property(tag, int(value))
                        except (TypeError, ValueError):
                            pass
                    elif tag in file.float_properties:
                        try:
                            file.set_property(tag, float(value))
                        except (TypeError, ValueError):
                            pass
                    else:
                        try:
                            value = value.strip()
                        except AttributeError:
                            pass
                        file.set_property(tag, value)

            self.extracted += 1
            self.apply_task.increment_progress(progress_step)

    def on_apply_done(self, *args):
        self.parent.toast_overlay.add_toast(
            Adw.Toast.new(
                ngettext(
                    # TRANSLATORS: {extracted} is a placeholder for the number
                    # of tracks the tags were succesfully extracted for.
                    # **Do not translate the text between the curly brackets!**
                    "Extracted tags for 1 track",
                    "Extracted tags for {extracted} tracks",
                    self.extracted,
                ).format(extracted=self.extracted)
            )
        )
        self._close()

    @Gtk.Template.Callback()
    def on_cancel(self, *args):
        if self.apply_task.is_running:
            self.apply_task.stop()
        self.preview_selector_button.disconnect(self._preview_update_conn)
        self.preview_selector_button.teardown()
        self.set_can_close(True)
        self._close()

    def _close(self):
        for conn in self._connections:
            self.disconnect(conn)
        self.pattern_entry.disconnect(self._entry_conn)
        self._connections = []
        self._entry_conn = None
        self.files = None
        self.set_can_close(True)
        self.close()

    def check_for_errors(self, *args):
        if self.custom_syntax_highlight.props.error:
            self.props.validation_passed = False
            return
        self.props.validation_passed = filename_valid(
            self.pattern_entry.get_text(), allow_path=False
        )



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

# Export some common functions:
from .misc import (  # noqa: F401
    all_equal,
    find_in_model,
    is_float,
    get_readable_length,
    inspect_prettyprint,
    title_case_preserve_uppercase,
    simplify_string,
    simplify_compare,
    reg_and_simple_cmp,
    file_is_sandboxed,
)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject
import asyncio
import struct

import mutagen.asf
from mutagen.id3 import PictureType

from .file import CoverType
from .file_mutagen_common import EartagFileMutagenCommon
from ..utils.misc import safe_int

# These are copied from the code for Quod Libet's wma handling:
KEY_TO_FRAME = {
    "album": "WM/AlbumTitle",
    "title": "Title",
    "artist": "Author",
    "albumartist": "WM/AlbumArtist",
    "composer": "WM/Composer",
    "lyricist": "WM/Writer",
    "conductor": "WM/Conductor",
    "remixer": "WM/ModifiedBy",
    "producer": "WM/Producer",
    "grouping": "WM/ContentGroupDescription",
    "discsubtitle": "WM/SubTitle",
    "tracknumber": "WM/TrackNumber",
    "totaltracknumber": "WM/TrackNumber",
    "discnumber": "WM/PartOfSet",
    "bpm": "WM/BeatsPerMinute",
    "copyright": "Copyright",
    "isrc": "WM/ISRC",
    "mood": "WM/Mood",
    "encodedby": "WM/EncodedBy",
    "musicbrainz_recordingid": "MusicBrainz/Track Id",
    "musicbrainz_trackid": "MusicBrainz/Release Track Id",
    "musicbrainz_albumid": "MusicBrainz/Album Id",
    "musicbrainz_artistid": "MusicBrainz/Artist Id",
    "musicbrainz_albumartistid": "MusicBrainz/Album Artist Id",
    "musicbrainz_trmid": "MusicBrainz/TRM Id",
    "musicip_puid": "MusicIP/PUID",
    "musicbrainz_releasegroupid": "MusicBrainz/Release Group Id",
    "releasedate": "WM/Year",
    "originalartist": "WM/OriginalArtist",
    "originalalbum": "WM/OriginalAlbumTitle",
    "albumsort": "WM/AlbumSortOrder",
    "artistsort": "WM/ArtistSortOrder",
    "albumartistsort": "WM/AlbumArtistSortOrder",
    "genre": "WM/Genre",
    "publisher": "WM/Publisher",
    "website": "WM/AuthorURL",
    "comment": "Description"
}  # fmt: skip


def unpack_image(data):
    """Unpacks an ASF/WMA picture."""
    # Picture type (this matches ID3 types)
    try:
        (picture_type, size) = struct.unpack_from("<bi", data)
    except struct.error:
        return None

    data = data[5:]

    # Mimetype string
    mime = b""
    while data:
        char, data = data[:2], data[2:]
        if char == b"\x00\x00":
            break
        mime += char
    else:
        return None
    mime = mime.decode("utf-16-le")

    # Description string
    description = b""
    while data:
        char, data = data[:2], data[2:]
        if char == b"\x00\x00":
            break
        description += char
    else:
        return None
    description = description.decode("utf-16-le")

    if size != len(data):
        return None

    return (mime, description, data, picture_type)


def pack_image(image_data, mime: str, image_type=3, description="thumbnail"):
    """Packs an image into ASF format."""
    size = len(image_data)

    data = struct.pack("<bi", image_type, size)
    data += mime.encode("utf-16-le") + b"\x00\x00"
    data += description.encode("utf-16-le") + b"\x00\x00"
    data += image_data

    return data


class EartagFileMutagenASF(EartagFileMutagenCommon):
    """EartagFile handler that uses mutagen for ASF support."""

    __gtype_name__ = "EartagFileMutagenASF"
    _supports_album_covers = True
    _cover_mimetypes = ["image/jpeg", "image/png"]
    _supports_full_dates = False

    # Copied from file.py, but excludes totaltracknumber, as ASF tags don't
    # have a way to specify it
    handled_properties = (
        "title",
        "artist",
        "album",
        "albumartist",
        "tracknumber",
        "genre",
        "releasedate",
        "comment",
    )

    supported_extra_tags = (
        "bpm", "composer", "copyright", "encodedby",
        "mood", "conductor", "discnumber", "publisher",
        "isrc", "discsubtitle",

        "albumartistsort", "albumsort", "artistsort",

        "musicbrainz_artistid", "musicbrainz_albumid",
        "musicbrainz_albumartistid", "musicbrainz_trackid",
        "musicbrainz_recordingid", "musicbrainz_releasegroupid"
    )  # fmt: skip

    async def load_from_file(self, path):
        self.mg_file = await asyncio.to_thread(mutagen.asf.ASF, path)
        await self.load_cover()
        self.setup_present_extra_tags()
        self.setup_original_values()

    def get_tag(self, tag_name):
        """Gets a tag's value using the KEY_TO_FRAME list as a guideline."""
        try:
            return str(self.mg_file.tags[KEY_TO_FRAME[tag_name.lower()]][0])
        except KeyError:
            return ""

    def set_tag(self, tag_name, value):
        """Sets a tag's value using the KEY_TO_FRAME list as a guideline."""
        frame_name = KEY_TO_FRAME[tag_name.lower()]

        # For float values that do not have numbers after the decimal point,
        # trim the trailing .0
        if tag_name in self.float_properties and value % 1 == 0:
            if value:
                stringified = str(safe_int(value))
            else:
                stringified = ""
        else:
            if value:
                stringified = str(value)
            else:
                stringified = ""

        self.mg_file.tags[frame_name] = [stringified]

    def has_tag(self, tag_name):
        """
        Returns True or False based on whether the tag with the given name is
        present in the file.
        """
        if tag_name == "totaltracknumber":
            return False
        if tag_name not in KEY_TO_FRAME:
            return False
        frame_name = KEY_TO_FRAME[tag_name.lower()]
        if frame_name in self.mg_file.tags:
            return True
        return False

    def delete_tag(self, tag_name):
        """Deletes the tag with the given name from the file."""
        frame_name = KEY_TO_FRAME[tag_name.lower()]
        if frame_name in self.mg_file.tags:
            del self.mg_file.tags[frame_name]
        self.mark_as_modified(tag_name, notify_prop=True)

    def delete_cover(self, cover_type: CoverType, clear_only=False):
        """Deletes the cover from the file."""
        if cover_type == CoverType.FRONT:
            pictypes = (PictureType.OTHER, PictureType.COVER_FRONT)
        elif cover_type == CoverType.BACK:
            pictypes = (PictureType.COVER_BACK,)
        else:
            raise ValueError

        pictures = []
        if "WM/Picture" in self.mg_file.tags:
            pictures = self.mg_file.tags["WM/Picture"]
        if pictures:
            for picture in pictures.copy():
                picture_type = unpack_image(picture.value)[3]
                if picture_type in pictypes:
                    pictures.remove(picture)
        self.mg_file.tags["WM/Picture"] = pictures

        if not clear_only:
            self._cleanup_cover(cover_type)

    def on_remove(self, *args):
        super().on_remove()

    async def set_cover_from_data(self, cover_type: CoverType, data: str, mime: str | None = None):
        if cover_type == CoverType.FRONT:
            pictype = PictureType.COVER_FRONT
        elif cover_type == CoverType.BACK:
            pictype = PictureType.COVER_BACK
        else:
            raise ValueError

        # Set cover in UI and check if it's valid
        ret = await self._set_cover_from_data(cover_type, data)
        if ret is False:
            return

        pictures = []
        if "WM/Picture" in self.mg_file.tags:
            pictures = self.mg_file.tags["WM/Picture"]

        packed_data = pack_image(data, mime, image_type=pictype)
        pictures.append(mutagen.asf.ASFValue(packed_data, mutagen.asf.BYTEARRAY))

        self.mg_file.tags["WM/Picture"] = pictures

    async def load_cover(self):
        """Loads the cover from the file and saves it to a temporary file."""
        if "WM/Picture" not in self.mg_file.tags:
            return None

        front_cover = None
        back_cover = None

        pictures = self.mg_file.tags["WM/Picture"]

        for picture in pictures:
            raw_data = picture.value
            mime, description, data, picture_type = unpack_image(raw_data)
            if picture_type == PictureType.COVER_FRONT and not front_cover:
                front_cover = data
            elif picture_type == PictureType.COVER_BACK and not back_cover:
                back_cover = data

        if front_cover:
            await self._set_cover_from_data(CoverType.FRONT, front_cover, modified=False)

        if back_cover:
            await self._set_cover_from_data(CoverType.BACK, back_cover, modified=False)

    @GObject.Property(type=str)
    def releasedate(self):
        return self.get_tag("releasedate")

    @releasedate.setter
    def releasedate(self, value):
        self.validate_date("releasedate", value)
        if value:
            if len(value) >= 4:
                try:
                    self.set_tag("releasedate", value[:4])
                except ValueError:
                    self.set_tag("releasedate", "")
            else:
                self.set_tag("releasedate", value)
            self.mark_as_modified("releasedate")
        else:
            self.delete_tag("releasedate")

    @GObject.Property(type=int)
    def totaltracknumber(self):
        return None

    @totaltracknumber.setter
    def totaltracknumber(self, value):
        return False



# SPDX-License-Identifier: MIT
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format="[%(asctime)s] eartag {%(pathname)s:%(lineno)d} %(levelname)s: %(message)s",
)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.events import GLibEventLoopPolicy

policy = GLibEventLoopPolicy()
event_loop = policy.get_event_loop()
event_loop.set_exception_handler(None)  # set default exception handler



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import Adw, Gdk, Gtk, GObject, Pango

from ..backends.file import CoverType
from .. import APP_GRESOURCE_PATH


class EartagPopoverButton(Gtk.Box):
    """
    Re-implementation of GtkMenuButton that doesn't have the same issues as
    it does. Notably:

    - doesn't prevent arrow navigation from working correctly
    - doesn't suffer from https://gitlab.gnome.org/GNOME/gtk/-/issues/5568
      (though that one is actually worked around in the AlbumCoverButton)
    """

    __gtype_name__ = "EartagPopoverButton"

    def __init__(self):
        super().__init__()
        self._popover = None
        self.toggle_button = Gtk.ToggleButton()
        self.append(self.toggle_button)
        self.props.accessible_role = Gtk.AccessibleRole.GROUP
        self.toggle_button.update_relation(
            (Gtk.AccessibleRelation.LABELLED_BY,),
            (Gtk.AccessibleList.new_from_list((self,)),),
        )

    @GObject.Property(type=Gtk.Widget)
    def child(self):
        return self.toggle_button.get_child()

    @child.setter
    def child(self, value):
        return self.toggle_button.set_child(value)

    @GObject.Property(type=Gtk.Popover)
    def popover(self):
        """The popover to display."""
        return self._popover

    @popover.setter
    def popover(self, value):
        if self._popover:
            self.remove(self._popover)
        self._popover = value
        self.toggle_button.bind_property(
            "active", self.popover, "visible", GObject.BindingFlags.BIDIRECTIONAL
        )
        self.append(self._popover)


def _delegate_getter(self, prop):
    return self.entry.get_delegate().get_property(prop)


def _delegate_setter(self, prop, value):
    return self.entry.get_delegate().set_property(prop, value)


class EartagEditableLabel(Gtk.Overlay, Gtk.Editable):
    """Editable label widget."""

    __gtype_name__ = "EartagEditableLabel"

    def __init__(self):
        super().__init__()
        self._last_edit_state = None

        self.add_css_class("editablelabel")

        self.entry = Gtk.Entry(valign=Gtk.Align.CENTER)
        self.entry.set_alignment(0.5)
        self.label = Gtk.Label(
            can_focus=False,
            can_target=False,
            wrap=True,
            valign=Gtk.Align.CENTER,
            lines=3,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            ellipsize=Pango.EllipsizeMode.MIDDLE,
            max_width_chars=128,
            justify=Gtk.Justification.CENTER,
            halign=Gtk.Align.CENTER,
        )
        self.label.set_cursor(Gdk.Cursor.new_from_name("text"))

        self.icon = Gtk.Image(
            icon_name="document-edit-symbolic",
            can_focus=False,
            can_target=False,
            receives_default=False,
            halign=Gtk.Align.END,
            valign=Gtk.Align.CENTER,
            accessible_role=Gtk.AccessibleRole.PRESENTATION,
        )
        self.icon.add_css_class("edit-icon")
        self.bind_property(
            "editing",
            self.icon,
            "visible",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.INVERT_BOOLEAN,
        )

        self.set_child(self.entry)
        self.add_overlay(self.label)
        self.add_overlay(self.icon)
        self.set_measure_overlay(self.label, True)

        self.entry.bind_property(
            "placeholder-text",
            self,
            "placeholder-text",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self.entry.get_delegate().bind_property(
            "has-focus", self, "editing", GObject.BindingFlags.SYNC_CREATE
        )

        self.entry.get_delegate().connect("changed", self.update_label)
        self.entry.connect("notify::placeholder-text", self.update_label)
        self.update_label()

        self.props.accessible_role = Gtk.AccessibleRole.GROUP

        self.connect("notify::editing", self.update_editing)
        self.update_editing()

        self.init_delegate()

    def update_editing(self, *args):
        if self._last_edit_state == self.props.editing:
            return

        self.entry.set_size_request(-1, self.label.get_height())

        if self.props.editing:
            # Switch which bit is visible
            self.label.props.visible = False
            self.entry.props.opacity = 1
        else:
            # Switch which bit is visible
            self.label.props.visible = True
            # self.label.props.opacity = 1
            self.entry.props.opacity = 0

        self.update_label()

        self._last_edit_state = self.props.editing

    def update_label(self, *args):
        if self.entry.props.text:
            self.label.set_text(self.entry.props.text)
            self.label.remove_css_class("dimmed")
        else:
            self.label.set_text(self.entry.props.placeholder_text or "")
            self.label.add_css_class("dimmed")

    def do_get_delegate(self):
        return self.entry.get_delegate()

    # Properties (TODO: find a better way to auto-proxy them?)
    cursor_position = GObject.Property(
        type=int,
        default=0,
        getter=lambda self: _delegate_getter(self, "cursor_position"),
        flags=GObject.ParamFlags.READABLE,
    )
    editable = GObject.Property(
        type=bool,
        default=True,
        getter=lambda self: _delegate_getter(self, "editable"),
        setter=lambda self, value: _delegate_setter(self, "editable", value),
    )
    enable_undo = GObject.Property(
        type=bool,
        default=True,
        getter=lambda self: _delegate_getter(self, "enable_undo"),
        setter=lambda self, value: _delegate_setter(self, "enable_undo", value),
    )
    max_width_chars = GObject.Property(
        type=int,
        default=-1,
        getter=lambda self: _delegate_getter(self, "max_width_chars"),
        setter=lambda self, value: _delegate_setter(self, "max_width_chars", value),
    )
    selection_bound = GObject.Property(
        type=int,
        default=0,
        getter=lambda self: _delegate_getter(self, "selection_bound"),
        setter=lambda self, value: _delegate_setter(self, "selection_bound", value),
    )
    text = GObject.Property(
        type=str,
        getter=lambda self: _delegate_getter(self, "text"),
        setter=lambda self, value: _delegate_setter(self, "text", value),
    )
    width_chars = GObject.Property(
        type=int,
        default=-1,
        getter=lambda self: _delegate_getter(self, "width_chars"),
        setter=lambda self, value: _delegate_setter(self, "width_chars", value),
    )
    xalign = GObject.Property(
        type=GObject.type_from_name("gfloat"),  # float converts to gdouble which causes warning
        default=0.0,
        getter=lambda self: _delegate_getter(self, "xalign"),
        setter=lambda self, value: _delegate_setter(self, "xalign", value),
    )

    # Custom properties
    editing = GObject.Property(type=bool, default=False)
    placeholder_text = GObject.Property(type=str)


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/albumcoverimage.ui")
class EartagAlbumCoverImage(Gtk.Stack):
    __gtype_name__ = "EartagAlbumCoverImage"

    no_cover = Gtk.Template.Child()
    cover_image = Gtk.Template.Child()

    file = None

    def __init__(self):
        super().__init__()
        self._cover_type = CoverType.FRONT
        self._connections = []
        self.connect("destroy", self.on_destroy)

    def on_destroy(self, *args):
        self.unbind_from_file()

    def bind_to_file(self, file):
        """
        Bind to a file.

        Note that only one file can be bound; binding to a file while a file is
        already bound will unbind the currently bound file.
        """
        if self.file:
            self.unbind_from_file()

        self.file = file

        if file.supports_album_covers:
            self.on_cover_change()
            self._connections.append(self.file.connect("cover-updated", self.on_cover_change))
        else:
            self.cover_image.set_from_file(None)
            self.on_cover_change()

    def unbind_from_file(self, file=None):
        if not self.file or (file and file != self.file):
            return False

        while len(self._connections) > 0:
            self.file.disconnect(self._connections[0])
            del self._connections[0]

        del self.file
        self.file = None

    def mark_as_empty(self):
        """In some cases, we need to force the cover to be shown as empty."""
        if self.get_visible_child() is not self.no_cover:
            self.set_visible_child(self.no_cover)
            self.notify("is-empty")

    def mark_as_nonempty(self):
        self.on_cover_change()

    @GObject.Property(type=bool, default=True)
    def is_empty(self):
        return self.get_visible_child() == self.no_cover

    @GObject.Signal
    def cover_changed(self):
        pass

    def on_cover_change(self, *args):
        if not self.file:
            self.mark_as_empty()
            return

        if self.cover_type == CoverType.FRONT:
            cover = self.file.front_cover
        elif self.cover_type == CoverType.BACK:
            cover = self.file.back_cover
        else:
            raise ValueError(self.cover_type)

        if cover:
            if self.get_visible_child() is not self.cover_image:
                self.set_visible_child(self.cover_image)
                self.notify("is-empty")

            if self.cover_image.get_pixel_size() <= 48:
                pixbuf = cover.cover_small
            else:
                pixbuf = cover.cover_large

            self.cover_image.set_from_pixbuf(pixbuf)
        else:
            self.mark_as_empty()

        self.emit("cover_changed")

    @GObject.Property(type=int, default=196)
    def pixel_size(self):
        return self.cover_image.get_pixel_size()

    @pixel_size.setter
    def pixel_size(self, value):
        self.cover_image.set_pixel_size(value)
        if value < 100:
            if value > 28:
                self.no_cover.set_pixel_size(24)
            else:
                self.no_cover.set_pixel_size(value - 4)
        else:
            self.no_cover.set_pixel_size(96)

    @GObject.Property(type=int)
    def cover_type(self):
        """Whether to display the front or back cover."""
        return self._cover_type

    @cover_type.setter
    def cover_type(self, value):
        self._cover_type = value
        self.on_cover_change()


class EartagModelExpanderRow(Adw.ExpanderRow):
    """
    Subclass of AdwExpanderRow that automatically fills rows based on
    a list model, and exposes some nice-to-have information about child
    rows.
    """

    def __init__(self):
        super().__init__()
        self.model = None

        # In today's episode of "fun facts about GTK/Adw internals":
        # did you know that an AdwExpanderRow consists of not one, but *two*
        # listboxes? That's right - the expander row appears to be a ListBox
        # of its own, and right below it is a GtkRevealer with *another*
        # ListBox in it, simply named "list".

        # Anyways, most of this isn't relevant here. What's relevant is that
        # this is, after all, a listbox inside, so we can just grab it with
        # get_template_child and use the model binding functions there.

        self.list = self.get_template_child(Adw.ExpanderRow, "list")

    def bind_model(self, *args, **kwargs):
        """See Gtk.ListBox.bind_model"""
        return self.list.bind_model(*args, **kwargs)

    def get_row_at_index(self, *args, **kwargs):
        """See Gtk.ListBox.get_row_at_index"""
        return self.list.get_row_at_index(*args, **kwargs)

    def set_placeholder(self, *args, **kwargs):
        """See Gtk.ListBox.set_placeholder"""
        return self.list.set_placeholder(*args, **kwargs)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GLib, GObject
import time
import traceback
import asyncio

from .._async import event_loop
from ..logger import logger
from typing import Coroutine


class EartagAsyncTask(GObject.Object):
    """
    Convenience class for creating async tasks.

    Provides a "progress" property that can be used by target functions
    to signify a progress change. This is a float from 0 to 1 and is
    passed directly to GtkProgressBar.

    The passed target function must be an async function/coroutine.
    """

    def __init__(self, target: Coroutine, *args, **kwargs):
        super().__init__()
        self._progress = 0
        self.target = target
        self.set_args(args, kwargs)
        self.task = None
        self.progress_lock = asyncio.Lock()

    def set_args(self, args=None, kwargs=None):
        if args:
            self.args = args
        else:
            self.args = []
        if kwargs:
            self.kwargs = kwargs
        else:
            self.kwargs = {}

    def wait_for_completion(self):
        while self.is_running:
            time.sleep(0.25)

    def stop(self):
        if self.task and not self.task.done():
            self.task.cancel()

    async def _run(self):
        try:
            return await self.target(*self.args, **self.kwargs)
        except Exception as e:
            # HACK: We do exception handling here since for some reason it
            # doesn't print the exceptions otherwise
            traceback.print_exc()
            raise e from e

    def run(self):
        self.task = event_loop.create_task(self._run())
        self.emit("task-started")
        self.task.add_done_callback(self.emit_task_done)

    @GObject.Property(type=float, minimum=0, maximum=1.0000001)
    def progress(self):
        """
        Float from 0 to 1 signifying the current progress of the operation.
        When the task is done, this automatically resets to 0.

        This value is set by the target function.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value

    @GObject.Signal
    def task_started(self):
        self.notify("is-running")

    @GObject.Signal
    def task_done(self):
        self.reset_progress()
        self.notify("is-running")

    @GObject.Property(type=bool, default=False)
    def is_running(self):
        if not self.task:
            return False
        return not self.task.done()

    def reset_progress(self):
        self.props.progress = 0

    def set_progress(self, value):
        """
        Wrapper around self.props.progress that updates the progress.
        """
        self.props.progress = value

    def increment_progress(self, value):
        """
        Wrapper around self.props.progress that increments the progress.
        """
        self.props.progress = self.props.progress + value

    async def set_progress_threadsafe(self, value):
        """
        Wrapper around self.props.progress that updates the progress,
        with a lock to prevent multiple threads incremeting it at the same time.
        """
        async with self.progress_lock:
            self.props.progress = value

    async def increment_progress_threadsafe(self, value):
        """
        Wrapper around self.props.progress that increments the progress,
        with a lock to prevent multiple threads incremeting it at the same time.
        """
        async with self.progress_lock:
            self.props.progress = self.props.progress + value

    progress_pulse = GObject.Signal()

    def emit_progress_pulse(self):
        """
        Send a progress pulse, for processes where the exact progress isn't known.
        """
        self.emit("progress-pulse")

    def emit_task_done(self, *args):
        """
        Wrapper around self.emit('task-done').
        """
        self.emit("task-done")


class EartagAsyncMultitasker(EartagAsyncTask):
    """
    Subclass of AsyncTask which runs a task using multiple workers with queued data.

    The Multitasker class provides a FIFO queue - this is a standard Python asyncio queue,
    and items can be picked up with the `.queue.get()` method.

    The typical loop of operations for the Multitasker looks like this:

    - Create the Multitasker object, set the target function
        - The target function must accept an EartagAsyncMultitasker object in the
          `tasker` parameter
    - In your run function:
        - Call the `.run()` method to spawn the workers
        - Start adding items to the queue with `.queue_put(item)`
        - Once all items are added, call `.queue_done()`
    - In your worker function:
        - Accept a parameter named "item"; this will contain the item to work on

    All AsyncTask properties/signals (like progress and task-done) apply here as well.
    task-done is only emitted once all workers finish execution.
    """

    # TODO: In the future (once Python 3.13 is actually wide-spread, so... maybe 2028?),
    # this code could be made nicer with Queue.shutdown(). Until then, our manual locking
    # code should do the job...

    def __init__(self, target: Coroutine, workers: int, *args, **kwargs):
        super().__init__(target, *args, **kwargs)
        assert target
        self.queue = asyncio.Queue()
        self.queue_done_event = asyncio.Event()
        self.workers = workers
        self.tasks = set()
        self._is_running = False
        self.running_lock = asyncio.Lock()

        #: If you want the multitasker to execute your task directly as the worker
        #: task, rather than managing the queue internally, set `.run_raw` to True.
        #: The "tasker" argument will be provided to your function with the Multitasker
        #: objects, so that you can write your own queue handling implementation.
        #: See _worker function for an example of how to use the APIs.
        self.run_raw: bool = False

        #: GLib priority to use for the worker tasks.
        self.priority: int = GLib.PRIORITY_DEFAULT_IDLE

        self.errors = []

        self.n_items = 0
        self.n_done = 0

    async def _worker(self):
        while True:
            item = await self.queue_get()
            if item is None:
                break
            try:
                await self.target(item, *self.args, **self.kwargs)
            except:  # noqa: E722
                self.errors.append(
                    f"{self.target}: Error while processing {item}:\n\n{traceback.format_exc()}"
                )
                logger.error(self.errors[-1])
                pass

            self.n_done += 1
            if self.queue_done_event.is_set():
                _progress_task = event_loop.create_task(
                    self.set_progress_threadsafe(self.n_done / self.n_items)
                )
                _progress_task.set_priority(GLib.PRIORITY_LOW)
            else:
                self.emit_progress_pulse()

    async def _run_multitasker(self):
        async with self.running_lock:
            self._is_running = True
            self.emit("task-started")
            async with asyncio.TaskGroup() as tg:
                for _i in range(self.workers):
                    _task = tg.create_task(self._worker())
                    _task.set_priority(self.priority)
                    self.tasks.add(_task)
            # The task group will block until all tasks are done
            self._is_running = False
            self.emit_task_done()

    def spawn_workers(self):
        """Start the task by spawning workers."""
        self.n_items = 0
        self.n_done = 0
        self.queue_done_event.clear()
        self.clear_errors()
        event_loop.create_task(self._run_multitasker())

    async def spawn_workers_async(self):
        """
        Version of spawn_workers for use with async functions (usually task
        groups).
        """
        self.n_items = 0
        self.n_done = 0
        self.queue_done_event.clear()
        self.clear_errors()
        await self._run_multitasker()

    def run(self):
        """Compatibility shim, please use spawn_workers() instead."""
        return self.spawn_workers()

    def stop(self):
        for task in self.tasks:
            if not task.done():
                task.cancel()

        del self.tasks
        self.tasks = set()

        self.queue_done_event.set()
        self.queue_done_event.clear()

    def wait_for_completion(self):
        while self.task and not self.task.done():
            time.sleep(0.25)

    async def wait_for_completion_async(self):
        async with self.running_lock:
            pass

    @GObject.Property(type=bool, default=False)
    def is_running(self):
        return self._is_running

    async def queue_get(self):
        """
        Wait for an item from the queue. Returns None if the queue is
        done being filled.
        """
        # Queue has items and we can just return something
        if not self.queue.empty():
            return self.queue.get_nowait()

        # Queue has no items, but it's done being filled, so we can return None
        elif self.queue_done_event.is_set():
            return None

        # Queue has no items but it's not done being filled; wait for the next
        # item or for the lock to be lifted
        queue_get_task = event_loop.create_task(self.queue.get())
        queue_done_task = event_loop.create_task(self.queue_done_event.wait())

        ## Wait for either the queue to return a new item, or for the queue to be done
        done, pending = await asyncio.wait(
            (queue_get_task, queue_done_task), return_when=asyncio.FIRST_COMPLETED
        )

        ## If the queue is marked as done, then return None
        if queue_done_task.done() and not queue_get_task.done():
            return None

        ## Otherwise get the item from the queue
        return queue_get_task.result()

    async def queue_put_async(self, item):
        """Put an item in the queue."""
        await self.queue.put(item)
        self.n_items += 1

    def queue_put(self, item):
        """Put an item in the queue."""
        event_loop.create_task(self.queue_put_async(item))

    async def queue_put_multiple_async(self, items, mark_as_done: bool = False):
        """Put all items from the list in the queue."""
        for item in items:
            await self.queue_put_async(item)
        if mark_as_done:
            self.queue_done()

    def queue_put_multiple(self, items, mark_as_done: bool = False):
        """
        Put an item in the queue.

        :param items: Iterable of items to put in the queue.
        :param mark_as_done: Mark the queue as done once finished.
        """
        event_loop.create_task(self.queue_put_multiple_async(items, mark_as_done=mark_as_done))

    def queue_done(self):
        """
        Indicate to the workers that the queue is done being filled.

        All currently blocked `queue_get` calls will be unblocked and return None.
        """
        self.queue_done_event.set()

    def clear_errors(self):
        """Clear error information from the tasker."""
        del self.errors
        self.errors = []



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject
import asyncio
import base64

import mutagen.mp4
from mutagen.mp4 import MP4Cover

from .file import CoverType
from .file_mutagen_common import EartagFileMutagenCommon
from ..utils.misc import safe_int
from ..utils.validation import get_mimetype_buffer

EMPTY_COVER = MP4Cover(
    base64.b64decode(
        "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAAC0lEQVQIW2NgAAIAAAUAAR4f7BQAAAAASUVORK5CYII="  # noqa: E501
    ),
    MP4Cover.FORMAT_PNG,
)


# These are copied from the code for Mutagen's EasyMP4 functions:
KEY_TO_FRAME = {
    "title": "\xa9nam",
    "album": "\xa9alb",
    "artist": "\xa9ART",
    "albumartist": "aART",
    "releasedate": "\xa9day",
    "comment": "\xa9cmt",
    "description": "desc",
    "grouping": "\xa9grp",
    "genre": "\xa9gen",
    "copyright": "cprt",
    "albumsort": "soal",
    "albumartistsort": "soaa",
    "artistsort": "soar",
    "titlesort": "sonm",
    "composersort": "soco",
    "tracknumber": "trkn",
    "totaltracknumber": "trkn",
    "cover": "covr",
    "bpm": "tmpo",
    "composer": "\xa9wrt",
    "encodedby": "\xa9enc",  # might be \xa9too
    "discnumber": "disk",

    "conductor": "----:com.apple.iTunes:CONDUCTOR",
    "discsubtitle": "----:com.apple.iTunes:DISCSUBTITLE",
    "language": "----:com.apple.iTunes:LANGUAGE",
    "mood": "----:com.apple.iTunes:MOOD",

    "musicbrainz_recordingid": "----:com.apple.iTunes:MusicBrainz Track Id",
    "musicbrainz_artistid": "----:com.apple.iTunes:MusicBrainz Artist Id",
    "musicbrainz_albumid": "----:com.apple.iTunes:MusicBrainz Album Id",
    "musicbrainz_albumartistid": "----:com.apple.iTunes:MusicBrainz Album Artist Id",
    "musicbrainz_releasegroupid": "----:com.apple.iTunes:MusicBrainz Release Group Id",
    "musicbrainz_trackid": "----:com.apple.iTunes:MusicBrainz Release Track Id",
}  # fmt: skip

# Annoyingly, there's no central MP4 tag standard; most software just tries to
# do what iTunes does (or did at some point... apparently they've had since 2005
# or even earlier to figure it out...)


class EartagFileMutagenMP4(EartagFileMutagenCommon):
    """EartagFile handler that uses mutagen for MP4 support."""

    __gtype_name__ = "EartagFileMutagenMP4"
    _supports_album_covers = True
    _cover_mimetypes = ["image/png"]
    _supports_full_dates = True

    supported_extra_tags = (
        "bpm", "composer", "copyright", "encodedby",
        "mood", "conductor", "discnumber", "language", "discsubtitle",

        "albumartistsort", "albumsort", "composersort", "artistsort",
        "titlesort",

        "musicbrainz_artistid", "musicbrainz_albumid",
        "musicbrainz_albumartistid", "musicbrainz_trackid",
        "musicbrainz_recordingid", "musicbrainz_releasegroupid"
    )  # fmt: skip

    async def load_from_file(self, path):
        self.mg_file = await asyncio.to_thread(mutagen.mp4.MP4, path)
        if self.mg_file.tags is None:
            self.mg_file.add_tags()
        await self.load_cover()
        self.setup_present_extra_tags()
        self.setup_original_values()

    def get_tag(self, tag_name):
        """Gets a tag's value using the KEY_TO_FRAME list as a guideline."""
        frame_name = KEY_TO_FRAME[tag_name.lower()]
        try:
            if frame_name.startswith("----"):
                return self.mg_file.tags[frame_name][0].decode("utf-8") or ""
            elif tag_name in self.int_properties:
                return self.mg_file.tags[frame_name][0]
            else:
                return self.mg_file.tags[frame_name][0] or ""
        except KeyError:
            return ""

    def set_tag(self, tag_name, value):
        """Sets a tag's value using the KEY_TO_FRAME list as a guideline."""
        frame_name = KEY_TO_FRAME[tag_name.lower()]
        if frame_name.startswith("----"):
            self.mg_file.tags[frame_name] = [value.encode("utf-8")]
        elif tag_name in ("bpm", "discnumber"):
            self.mg_file.tags[frame_name] = [safe_int(value)]
        else:
            self.mg_file.tags[frame_name] = [str(value)]

    def has_tag(self, tag_name):
        """
        Returns True or False based on whether the tag with the given name is
        present in the file.
        """
        if tag_name == "totaltracknumber":
            return bool(self.totaltracknumber)
        if tag_name not in KEY_TO_FRAME:
            return False
        frame_name = KEY_TO_FRAME[tag_name.lower()]
        if frame_name in self.mg_file.tags:
            return True
        return False

    def delete_tag(self, tag_name):
        """Deletes the tag with the given name from the file."""
        frame_name = KEY_TO_FRAME[tag_name.lower()]
        if tag_name.lower() == "releasedate":
            self._releasedate_cached = ""
        if frame_name in self.mg_file.tags:
            del self.mg_file.tags[frame_name]
        self.mark_as_modified(tag_name, notify_prop=True)

    def delete_cover(self, cover_type: CoverType, clear_only=False):
        """Deletes the cover from the file."""
        if "covr" not in self.mg_file.tags:
            if not clear_only:
                self._cleanup_cover(cover_type)
            return

        cover_val = self.mg_file.tags["covr"]
        if not cover_val:
            if not clear_only:
                self._cleanup_cover(cover_type)
            return

        if cover_type == CoverType.FRONT:
            if len(cover_val) == 1:
                del self.mg_file.tags["covr"]
                if not clear_only:
                    self._cleanup_cover(cover_type)
                return
            elif len(cover_val) >= 2:
                cover_val[0] = EMPTY_COVER

        elif cover_type == CoverType.BACK:
            if len(cover_val) == 1:
                if not clear_only:
                    self._cleanup_cover(cover_type)
                return
            elif len(cover_val) == 2:
                cover_val = [cover_val[0]]
            elif len(cover_val) > 2:
                cover_val[1] = EMPTY_COVER

        try:
            if cover_val[0] == EMPTY_COVER and cover_val[1] == EMPTY_COVER:
                cover_val = []
        except IndexError:
            pass

        self.mg_file.tags["covr"] = cover_val

        if not clear_only:
            self._cleanup_cover(cover_type)

    def on_remove(self, *args):
        super().on_remove()

    async def set_cover_from_data(self, cover_type: CoverType, data: str, mime: str | None = None):
        if not mime:
            mime = get_mimetype_buffer(data)

        # Set cover in UI and check if it's valid
        ret = await self._set_cover_from_data(cover_type, data)
        if ret is False:
            return

        cover_val = []
        if "covr" in self.mg_file.tags:
            cover_val = self.mg_file.tags["covr"]

        if cover_type == CoverType.FRONT:
            try:
                cover_val[0] = MP4Cover(data, MP4Cover.FORMAT_PNG)
            except IndexError:
                cover_val.append(MP4Cover(data, MP4Cover.FORMAT_PNG))
            self.mg_file.tags["covr"] = cover_val

        elif cover_type == CoverType.BACK:
            # I can't figure out how to specify a back cover in MP4 tags -
            # I found one forum post on a closed-source tagger's forum that
            # claims MP4 supports front and back covers, but I can't find
            # anything in Mutagen that would allow me to specify which one
            # to use... so we just set the second cover in the cover list
            # instead.
            if len(cover_val) == 0:
                # In general, this should never happen... if someone is setting
                # a back cover, then they probably set a front cover as well,
                # but let's handle this just in case.
                #
                # Set the front cover to a transparent 1x1 pixel.
                cover_val = [EMPTY_COVER, MP4Cover(data, MP4Cover.FORMAT_PNG)]
            elif len(cover_val) == 1:
                cover_val.append(MP4Cover(data, MP4Cover.FORMAT_PNG))
            elif len(cover_val) >= 2:
                cover_val[1] = MP4Cover(data, MP4Cover.FORMAT_PNG)

            self.mg_file.tags["covr"] = cover_val

    async def load_cover(self):
        """Loads the cover from the file and saves it to a temporary file."""
        if "covr" not in self.mg_file.tags or not self.mg_file.tags["covr"]:
            self._front_cover_path = None
            self._back_cover_path = None
            return None

        picture = self.mg_file.tags["covr"][0]
        if picture != EMPTY_COVER:
            await self._set_cover_from_data(CoverType.FRONT, picture, modified=False)

        try:
            picture_back = self.mg_file.tags["covr"][1]
            assert picture_back != EMPTY_COVER
        except (AssertionError, IndexError, KeyError):
            pass
        else:
            await self._set_cover_from_data(CoverType.BACK, picture_back, modified=False)

    @GObject.Property(type=str)
    def releasedate(self):
        if not self._releasedate_cached:
            self._releasedate_cached = self.get_tag("releasedate")
        return self._releasedate_cached

    @releasedate.setter
    def releasedate(self, value):
        if value:
            self.validate_date("releasedate", value)
            self._releasedate_cached = value
            if "releasedate" not in self._error_fields:
                self.set_tag("releasedate", value)
        else:
            self.delete_tag("releasedate")
        self.mark_as_modified("releasedate")

    @GObject.Property(type=int)
    def tracknumber(self):
        if "trkn" not in self.mg_file.tags:
            return None

        return safe_int(self.mg_file.tags["trkn"][0][0])

    @tracknumber.setter
    def tracknumber(self, value):
        if safe_int(value) == -1 or not value:
            value = 0
        if self.totaltracknumber:
            self.mg_file.tags["trkn"] = [(safe_int(value), safe_int(self.totaltracknumber))]
        else:
            if value:
                self.mg_file.tags["trkn"] = [(safe_int(value), 0)]
            elif self.has_tag("tracknumber"):
                self.delete_tag("tracknumber")
        self.mark_as_modified("tracknumber")

    @GObject.Property(type=int)
    def totaltracknumber(self):
        if "trkn" not in self.mg_file.tags:
            return None

        tracknum_raw = self.mg_file.tags["trkn"][0]
        if len(tracknum_raw) > 1:
            return safe_int(tracknum_raw[1])
        return None

    @totaltracknumber.setter
    def totaltracknumber(self, value):
        if safe_int(value) == -1 or not value:
            value = 0

        if self.tracknumber:
            self.mg_file.tags["trkn"] = [(safe_int(self.tracknumber), safe_int(value))]
        else:
            if value:
                self.mg_file.tags["trkn"] = [(0, safe_int(value))]
            elif self.has_tag("tracknumber"):
                self.delete_tag("tracknumber")
        self.mark_as_modified("totaltracknumber")

    @GObject.Property(type=int)
    def discnumber(self):
        if "disk" not in self.mg_file.tags and "DISK" not in self.mg_file.tags:
            return None

        try:
            return safe_int(self.mg_file.tags["disk"][0][0])
        except KeyError:
            return safe_int(self.mg_file.tags["DISK"][0][0])

    @discnumber.setter
    def discnumber(self, value):
        if safe_int(value) == -1 or not value:
            value = 0

        if value:
            self.mg_file.tags["disk"] = [(safe_int(value), 0)]
        elif self.has_tag("discnumber"):
            self.delete_tag("discnumber")
        self.mark_as_modified("discnumber")



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

import aiofiles
import aiofiles.os
import os.path
import filetype
from filetype.types import AUDIO as audio_matchers
from filetype.types import IMAGE as image_matchers
from filetype.types import VIDEO as video_matchers
from filetype.types.base import Type
import mimetypes

from collections.abc import Iterable
from typing import Union, Optional

VALID_AUDIO_MIMES = (
    "application/ogg",
    "application/x-ogg",
    "audio/aac",
    "audio/flac",
    "audio/mp3",
    "audio/mp4",
    "audio/mpeg",
    "audio/ogg",
    "audio/wav",
    "audio/x-flac",
    "audio/x-m4a",
    "audio/x-mp3",
    "audio/x-mpeg",
    "audio/x-ms-wma",
    "audio/x-vorbis+ogg",
    "audio/x-wav",
    "video/mp4",
    "video/x-ms-asf",
    "video/x-wmv",
)

VALID_IMAGE_MIMES = (
    # Supported natively by most formats:
    "image/jpg",
    "image/jpeg",
    "image/png",
    # Converted using Pillow:
    "image/bmp",
    "image/jp2",
    "image/webp",
)

# Custom filetype.py filetypes for formats it doesn't usually support


class ImprovedMP3(Type):
    """
    Improved MP3 filetype from latest main branch;
    see https://github.com/h2non/filetype.py/blob/master/filetype/types/audio.py
    """

    MIME = "audio/mpeg"
    EXTENSION = "mp3"

    def __init__(self):
        super(ImprovedMP3, self).__init__(mime=ImprovedMP3.MIME, extension=ImprovedMP3.EXTENSION)

    def match(self, buf):
        if len(buf) > 2:
            if buf[0] == 0x49 and buf[1] == 0x44 and buf[2] == 0x33:
                return True

            if buf[0] == 0xFF:
                if (
                    buf[1] == 0xE2  # MPEG 2.5 with error protection
                    or buf[1] == 0xE3  # MPEG 2.5 w/o error protection
                    or buf[1] == 0xF2  # MPEG 2 with error protection
                    or buf[1] == 0xF3  # MPEG 2 w/o error protection
                    or buf[1] == 0xFA  # MPEG 1 with error protection
                    or buf[1] == 0xFB  # MPEG 1 w/o error protection
                ):
                    return True
        return False


class Wma(Type):
    """Implements the WMA audio type matcher."""

    MIME = "audio/x-ms-wma"
    EXTENSION = "wma"

    def __init__(self):
        super(Wma, self).__init__(mime=Wma.MIME, extension=Wma.EXTENSION)

    def match(self, buf):
        if len(buf) > 14:
            if (
                buf[0] == 0x30
                and buf[1] == 0x26
                and buf[2] == 0xB2
                and buf[3] == 0x75
                and buf[4] == 0x8E
                and buf[5] == 0x66
                and buf[6] == 0xCF
                and buf[7] == 0x11
                and buf[8] == 0xA6
                and buf[9] == 0xD9
                and buf[10] == 0x00
                and buf[11] == 0xAA
                and buf[12] == 0x00
                and buf[13] == 0x62
                and buf[14] == 0xCE
                and buf[15] == 0x6C
            ):
                return True
        return False


CUSTOM_MATCHERS = (ImprovedMP3(), Wma())
FILETYPE_MATCHERS = CUSTOM_MATCHERS + audio_matchers + image_matchers + video_matchers

_mimetype_cache = {}


def get_mimetype_buffer(data):
    """Get mimetype from buffer."""
    mimetype = filetype.match(data, matchers=FILETYPE_MATCHERS)
    if mimetype:
        return mimetype.mime
    return None


def get_mimetype(
    path: Union[str, os.PathLike],
    no_extension_guess: bool = False,
    no_cache: bool = False,
) -> Optional[str]:
    """
    Return the mimetype (or None) for the file with the given path or data.

    :param path: Path to the file (or buffer data).
    :param no_extension_guess: If True, skips guessing the filetype from the extension.
    :param no_cache: Skip the mimetype cache.
    """
    if not no_cache:
        if path in _mimetype_cache:
            return _mimetype_cache[path]

    mimetype = filetype.match(path, matchers=FILETYPE_MATCHERS)

    ret = None
    if mimetype:
        ret = mimetype.mime
    elif not no_extension_guess:
        # Try to guess mimetype from file extension if filetype match fails
        guess = mimetypes.guess_type(path)
        if guess and guess[0]:
            ret = guess[0]

    if not no_cache:
        _mimetype_cache[path] = ret

    return ret


async def get_mimetype_async(
    path: Union[str, os.PathLike],
    no_extension_guess: bool = False,
    no_cache: bool = False,
) -> Optional[str]:
    """
    Return the mimetype (or None) for the file with the given path or data.

    Async-safe version of get_mimetype.

    :param path: Path to the file (or buffer data).
    :param no_extension_guess: If True, skips guessing the filetype from the extension.
    :param no_cache: Skip the mimetype cache.
    """
    ret = None

    if not no_cache:
        if path in _mimetype_cache:
            return _mimetype_cache[path]

    async with aiofiles.open(path, "rb") as file:
        ret = get_mimetype_buffer(await file.read(2048))

    if not ret and not no_extension_guess:
        # Try to guess mimetype from file extension if filetype match fails
        guess = mimetypes.guess_type(path)
        if guess and guess[0]:
            ret = guess[0]

    if not no_cache:
        _mimetype_cache[path] = ret

    return ret


def is_valid_file(
    path: Union[str, os.PathLike],
    valid_mime_types: Iterable[str],
    no_cache: bool = False,
):
    """
    Takes a path to a file and returns True if it's supported, False otherwise.
    """
    # In Flatpak, some files don't exist; double-check to make sure
    if not os.path.exists(path):
        return False

    mimetype = get_mimetype(path, no_cache=no_cache)

    if not mimetype or mimetype not in valid_mime_types:
        return False
    return True


def is_valid_music_file(path: Union[str, os.PathLike], no_cache: bool = False):
    """Check if the file at the provided path is a supported audio file."""
    return is_valid_file(path, VALID_AUDIO_MIMES, no_cache=no_cache)


def is_valid_image_file(path: Union[str, os.PathLike], no_cache: bool = False):
    """Check if the file at the provided path is a supported image file."""
    return is_valid_file(path, VALID_IMAGE_MIMES, no_cache=no_cache)


async def is_valid_file_async(
    path: Union[str, os.PathLike],
    valid_mime_types: Iterable[str],
    no_cache: bool = False,
):
    """
    Takes a path to a file and returns True if it's supported, False otherwise.

    Async version of is_valid_file.
    """
    # In Flatpak, some files don't exist; double-check to make sure
    if not await aiofiles.os.path.exists(path):
        return False

    mimetype = await get_mimetype_async(path, no_cache=no_cache)

    if not mimetype or mimetype not in valid_mime_types:
        return False
    return True


async def is_valid_music_file_async(path: Union[str, os.PathLike], no_cache: bool = False):
    """Check if the file at the provided path is a supported audio file."""
    return await is_valid_file_async(path, VALID_AUDIO_MIMES, no_cache=no_cache)


async def is_valid_image_file_async(path: Union[str, os.PathLike], no_cache: bool = False):
    """Check if the file at the provided path is a supported image file."""
    return await is_valid_file_async(path, VALID_IMAGE_MIMES, no_cache=no_cache)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from ..backends.file import VALID_TAGS

import os.path
import re

EXTRACTABLE_TAGS = VALID_TAGS + ("length", "bitrate")


def extract_tags_from_filename(
    filename: str, placeholder: str, positions: bool = False, strip_common_suffixes: bool = False
) -> dict:
    """
    Takes a filename and a placeholder string and splits the filename
    up into a dict containing tag data.

    :params filename: Filename or path to file for extraction. The basename and
        extension are split off inside this function.
    :params placeholder: The placeholder string for tag extraction.
    :params positions: If True, instead of returning just the tag contents
        in the results dict, returns tuples containing the value as the first
        element and the position span as the second.
    :params strip_common_suffixes: If True, removes common suffixes found in
        filenames of downloaded files
    """
    filename = os.path.splitext(os.path.basename(filename))[0]

    # Step 0. Strip common suffixes.
    if strip_common_suffixes:
        # Modern yt-dlp: square brackets with ID inside
        # (could be YouTube ID, or longer for e.g. SoundCloud,
        # so we don't limit it)
        if filename.endswith("]"):
            try:
                filename_stripped = re.match(r"(.*?) \[(.*)\]", filename).group(1)
                assert filename_stripped
            except (AssertionError, AttributeError, IndexError):
                pass
            else:
                filename = filename_stripped

        # Old youtube-dl: "-" and then YouTube ID. To minimize
        # the likelihood for misdetections, we only check for
        # YouTube IDs that have numbers or special characters
        # in them.
        try:
            if re.match(r"-([A-Za-z0-9_\-]{11})", filename[-12:]) and re.search(
                r"[0-9_\-]", filename[-11:]
            ):
                filename = filename[:-12]
        except IndexError:
            pass

    # Step 1. Split placeholder string into static strings and placeholders.
    placeholder_split = [x for x in re.split("({.*?})", placeholder) if x]

    tags = []

    # Step 2. Generate regex rule from the placeholders. This replaces every
    # valid placeholder found with a Regex capture group.
    pattern = "^"
    for element in placeholder_split:
        if element.startswith("{") and element.endswith("}"):
            tag = element[1:-1]
            if tag in EXTRACTABLE_TAGS and tag not in tags:
                pattern += f"(?P<{tag}>.*?)"
                tags.append(tag)
                continue
        pattern += re.escape(element)
    pattern += "$"

    # Pango attributes (used for syntax highlighting) use offsets calculated
    # in bytes, not Python characters, so we encode the pattern and filename
    # to UTF-8 so that the returned group positions match the byte count.
    pattern = pattern.encode("utf-8")
    filename = filename.encode("utf-8")

    match = re.match(pattern, filename)

    if not match:
        return {}

    out = {}

    for tag in tags:
        tag_matched = match.group(tag)
        if tag_matched:
            tag_value = tag_matched.decode("utf-8")
            if positions:
                out[tag] = (tag_value, match.span(tag))
            else:
                out[tag] = tag_value

    return out



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from .file_mutagen_asf import EartagFileMutagenASF  # noqa: F401
from .file_mutagen_id3 import EartagFileMutagenID3  # noqa: F401
from .file_mutagen_mp4 import EartagFileMutagenMP4  # noqa: F401
from .file_mutagen_vorbis import EartagFileMutagenVorbis  # noqa: F401


from .file import BASIC_TAGS, EXTRA_TAGS, TAG_NAMES  # noqa: F401

FILE_CLASSES = (
    EartagFileMutagenASF,
    EartagFileMutagenID3,
    EartagFileMutagenMP4,
    EartagFileMutagenVorbis,
)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject, Gtk
import re

from .misc import is_float


class EartagEntryLimiters(GObject.Object):
    """
    Common input validators for entries. Assumes inheriting object is a
    descendant of GtkEditable (and thus has a get_delegate method and
    insert-text signal).

    Adds properties that can be quickly set/unset to connect/disconnect
    an entry.

    You might be able to use multiple limiters at once, but this is
    completely untested, and you will have to re-set the input purpose
    manually since every property setter overrides it.
    """

    def setup_limiters(self):
        """
        Call this **AFTER** the super().__init__() call in the inheritant.
        """
        self._limiter_connections = {}
        self._limiter_connections["destroy"] = self.connect(
            "destroy", self._break_limiter_connections
        )

    def _break_limiter_connections(self, *args):
        for conn_type in ("numeric", "float", "date"):
            if conn_type in self._limiter_connections:
                self.get_delegate().disconnect(self._limiter_connections[conn_type])
        self.disconnect(self._limiter_connections["destroy"])
        self._limiter_connections = {}

    # The following properties should be copied into the subclass
    # (see https://gitlab.gnome.org/GNOME/pygobject/-/issues/577)
    """
    @GObject.Property(type=bool, default=False)
    def is_numeric(self):
        return self._is_numeric_prop

    @is_numeric.setter
    def is_numeric(self, value):
        self._is_numeric_prop = value

    @GObject.Property(type=bool, default=False)
    def is_float(self):
        return self._is_float_prop

    @is_float.setter
    def is_float(self, value):
        self._is_float_prop = value

    @GObject.Property(type=bool, default=False)
    def is_date(self):
        return self._is_date_prop

    @is_date.setter
    def is_date(self, value):
        self._is_date_prop = value
    """

    #
    # Numeric validator: allows only digits ([0-9]).
    #

    @property
    def _is_numeric_prop(self):
        try:
            return self._is_numeric
        except AttributeError:
            self._is_numeric = False
            return False

    @_is_numeric_prop.setter
    def _is_numeric_prop(self, value):
        try:
            if value == self._is_numeric:
                return
        except AttributeError:
            pass

        self._is_numeric = value
        if value:
            self.set_input_purpose(Gtk.InputPurpose.DIGITS)
            self._limiter_connections["numeric"] = self.get_delegate().connect(
                "insert-text", EartagEntryLimiters.disallow_nonnumeric
            )
        else:
            self.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
            if "numeric" in self._limiter_connections:
                self.get_delegate().disconnect(self._limiter_connections["numeric"])
                del self._limiter_connections["numeric"]

    @staticmethod
    def disallow_nonnumeric(entry, text, length, position, *args):
        if not text:
            return
        if not text.isdigit():
            GObject.signal_stop_emission_by_name(entry, "insert-text")

    #
    # Float validator: allows digits and one dot.
    #

    @property
    def _is_float_prop(self):
        try:
            return self._is_float
        except AttributeError:
            self._is_float = False
            return False

    @_is_float_prop.setter
    def _is_float_prop(self, value):
        try:
            if value == self._is_float:
                return
        except AttributeError:
            pass

        self._is_float = value
        if value:
            self.set_input_purpose(Gtk.InputPurpose.NUMBER)
            self._limiter_connections["float"] = self.get_delegate().connect(
                "insert-text", EartagEntryLimiters.disallow_nonfloat
            )
        else:
            self.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
            if "float" in self._limiter_connections:
                self.get_delegate().disconnect(self._limiter_connections["float"])
                del self._limiter_connections["float"]

    @staticmethod
    def disallow_nonfloat(entry, text, length, position, *args):
        if not text:
            return
        if "." in text and "." in entry.get_text():
            GObject.signal_stop_emission_by_name(entry, "insert-text")
        if text != "." and not is_float(text):
            GObject.signal_stop_emission_by_name(entry, "insert-text")

    #
    # Date validator: allows YYYYYYYYY..., YYYY-MM or YYYY-MM-DD dates.
    #

    @property
    def _is_date_prop(self):
        try:
            return self._is_date
        except AttributeError:
            self._is_date = False
            return False

    @_is_date_prop.setter
    def _is_date_prop(self, value):
        try:
            if value == self._is_date:
                return
        except AttributeError:
            pass

        self._is_date = value
        self.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
        if value:
            self._limiter_connections["date"] = self.get_delegate().connect(
                "insert-text", EartagEntryLimiters.disallow_nondate
            )
        else:
            if "date" in self._limiter_connections:
                self.get_delegate().disconnect(self._limiter_connections["date"])
                del self._limiter_connections["date"]

    @staticmethod
    def disallow_nondate(entry, text, length, position, *args):
        if not text:
            return
        elif not re.match("^[0-9-]*$", text):
            GObject.signal_stop_emission_by_name(entry, "insert-text")
            return

        current_text = entry.get_buffer().get_text()

        current_length = len(current_text)
        if current_length + length > 10:
            GObject.signal_stop_emission_by_name(entry, "insert-text")
            return

        sep_count = (current_text + text).count("-")

        if sep_count > 2:
            GObject.signal_stop_emission_by_name(entry, "insert-text")
            return



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from . import APP_GRESOURCE_PATH
from .backends.file import EartagFile, VALID_TAGS, TAG_NAMES
from .config import config
from .utils import get_readable_length, file_is_sandboxed
from .utils.misc import filename_valid
from .utils.tagselector import EartagTagSelectorButton  # noqa: F401
from .utils.tagsyntaxhighlight import (
    EartagPlaceholderSyntaxHighlighter,
    attr_foreground_new,
    THEMES,
)
from .utils.previewselector import EartagPreviewSelectorButton  # noqa: F401

from gi.repository import Adw, GLib, Gtk, Gio, GObject, Pango
import os
import re


def tag_is_int(file, tag):
    return tag in file.int_properties + file.float_properties + ("length", "bitrate")


def get_formatted_tag(file: "EartagFile", tag: str) -> str:
    """
    Returns the value of the tag formatted according to placeholder preview rules.
    """
    parsed_value = ""

    if tag in VALID_TAGS + ("length", "bitrate"):
        value = file.get_property(tag)
        if not value:
            if tag == "title":
                # TRANSLATORS: Default placeholder string for a missing title
                # tag, used when renaming files - if the title tag is not
                # present in the file, this placeholder will be used instead.
                parsed_value = _("Untitled")
            elif tag_is_int(file, tag):
                if tag.endswith("tracknumber") or tag.endswith("discnumber"):
                    parsed_value = "00"
                else:
                    parsed_value = "0"
            else:
                # TRANSLATORS: Default placeholder string for missing tags,
                # used when renaming files - if a tag is not present in the
                # file, this placeholder will be used instead.
                # {tag_name} is a placeholder for the human-readable tag name.
                # **Do not change the text between the curly brackets!**
                parsed_value = _("Unknown {tag_name}").format(tag_name=TAG_NAMES[tag])

        else:
            if tag_is_int(file, tag):
                if not value or value < 0:
                    value = 0
                if tag == "length":
                    parsed_value = get_readable_length(int(value))
                elif tag.endswith("tracknumber") or tag.endswith("discnumber"):
                    parsed_value = str(value).zfill(2)
                else:
                    parsed_value = str(value)
            else:
                parsed_value = str(value)
                # Other characters are cleaned by cleanup_filename; these are
                # potential file path separators we need to filter out first,
                # otherwise they'll get treated as folders (valid as far as
                # renaming and moving to a folder is concerned).
                parsed_value = parsed_value.replace("/", "_")

    return parsed_value


def parse_placeholder_string(placeholder: str, file: "EartagFile", positions: bool = False) -> dict:
    """
    Takes a placeholder string and a file and returns a string filled with the
    placeholders.
    """
    if placeholder == "":
        if positions:
            return "", []
        return ""

    # Pango attributes (used for syntax highlighting) use offsets calculated
    # in bytes, not Python characters, so we encode the placeholder to UTF-8
    # so that the returned group positions match the byte count.
    placeholder = placeholder.encode("utf-8")

    n = 0
    offset = 0
    out = placeholder
    _positions = []
    for match in re.finditer(r"{.*?}".encode("utf-8"), placeholder):
        try:
            tag_name = match.group(0)[1:-1].decode("utf-8")
        except (IndexError, UnicodeDecodeError):
            continue

        if "{" in tag_name or "}" in tag_name:
            continue

        if tag_name == "":
            continue

        if tag_name not in VALID_TAGS + ("length", "bitrate"):
            continue

        formatted_value = get_formatted_tag(file, tag_name).encode("utf-8")
        out = out.replace(("{" + tag_name + "}").encode("utf-8"), formatted_value, 1)

        _positions.append(
            (
                match.span(0)[0] + offset,
                match.span(0)[0] + offset + len(formatted_value),
            )
        )
        offset += len(formatted_value) - len(("{" + tag_name + "}").encode("utf-8"))
        n += 1

    out = out.decode("utf-8")

    if positions:
        return out, _positions
    return out


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/rename.ui")
class EartagRenameDialog(Adw.Dialog):
    __gtype_name__ = "EartagRenameDialog"

    toast_overlay = Gtk.Template.Child()

    rename_button = Gtk.Template.Child()
    rename_progress = Gtk.Template.Child()
    error_banner = Gtk.Template.Child()

    sandbox_warning_banner = Gtk.Template.Child()

    filename_entry = Gtk.Template.Child()
    preview_entry = Gtk.Template.Child()
    preview_selector_button = Gtk.Template.Child()

    _last_folder = None

    folder_selector_row = Gtk.Template.Child()
    folder_remove_button = Gtk.Template.Child()

    validation_passed = GObject.Property(type=bool, default=True)

    tag_selector = Gtk.Template.Child()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.file_manager = parent.file_manager
        self._folder = None
        self._has_sandboxed_files = False
        self._renamed = 0

        self.files = self.file_manager.selected_files_list.copy()
        for file in self.files:
            if file_is_sandboxed(file.path):
                self._has_sandboxed_files = True
                break

        self.syntax_highlight = EartagPlaceholderSyntaxHighlighter(
            self.filename_entry, "entry", allow_duplicates=True
        )

        self.folder_chooser = Gtk.FileDialog(modal=True)
        self.bind_property(
            "folder",
            self.folder_selector_row,
            "subtitle",
            GObject.BindingFlags.SYNC_CREATE,
        )
        if EartagRenameDialog._last_folder is not None:
            self.props.folder = EartagRenameDialog._last_folder
        elif os.path.exists(config["rename-base-folder"]):
            self.props.folder = config["rename-base-folder"]

        self.file_manager.rename_task.bind_property("progress", self.rename_progress, "fraction")
        self.file_manager.rename_task.connect("task-done", self.on_done)

        config.bind(
            "rename-placeholder",
            self.filename_entry,
            "text",
            Gio.SettingsBindFlags.DEFAULT,
        )

        if self._has_sandboxed_files and not self.props.folder:
            self.sandbox_warning_banner.props.revealed = True

        self.preview_selector_button.set_files(self.files)
        self.preview_selector_button.set_formatting_function(self.generate_preview_attrs)
        self._preview_update_conn = self.preview_selector_button.connect(
            "notify::selected-index", self.update_preview
        )

        self.connect("notify::folder", self.validate_placeholder)
        self.connect("notify::validation-passed", self.update_rename_button_sensitivity)
        self.validate_placeholder()
        self.update_rename_button_sensitivity()

    def validate_placeholder(self, *args):
        """Validates the filename input."""
        placeholder = self.filename_entry.get_text()
        if "/" in placeholder and not self.props.folder:
            self.props.validation_passed = False
        else:
            if self.props.folder and placeholder.startswith("/"):
                self.props.validation_passed = False
            else:
                self.props.validation_passed = filename_valid(
                    self.filename_entry.get_text(), allow_path=bool(self.props.folder)
                )
        self.update_rename_button_sensitivity()
        self.folder_remove_button.props.sensitive = bool(self.props.folder)

    def update_rename_button_sensitivity(self, *args):
        if self.props.validation_passed:
            self.filename_entry.remove_css_class("error")
        else:
            self.filename_entry.add_css_class("error")

        has_sandboxed = self._has_sandboxed_files and not self.props.folder

        self.rename_button.set_sensitive(self.props.validation_passed and not has_sandboxed)

    @Gtk.Template.Callback()
    def add_placeholder_from_selector(self, selector, tag, *args):
        """Adds a new placeholder based on the tag selector."""
        self.filename_entry.insert_text("{" + tag + "}", self.filename_entry.props.cursor_position)

    # Move to folder feature

    @GObject.Property(type=str, default=None)
    def folder(self):
        """Base folder to use for the "move to folder" option."""
        return self._folder

    @folder.setter
    def folder(self, value):
        self._folder = value
        EartagRenameDialog._last_folder = value
        if not value:
            config["rename-base-folder"] = ""
        elif not value.startswith("/run/user/"):
            config["rename-base-folder"] = value

        self.sandbox_warning_banner.props.revealed = self._has_sandboxed_files and not value

    @Gtk.Template.Callback()
    def show_folder_selector(self, *args):
        self.folder_chooser.select_folder(self.parent, None, self.select_folder_from_selector, None)

    def select_folder_from_selector(self, source, result, data):
        try:
            response = self.folder_chooser.select_folder_finish(result)
        except GLib.GError:
            return

        if not response:
            return

        folder = response.get_path()
        try:
            assert os.path.exists(folder)
        except AssertionError:
            self.toast_overlay.add_toast(Adw.Toast.new(_("Selected folder does not exist")))
        else:
            try:
                assert os.access(folder, os.W_OK)
            except AssertionError:
                self.toast_overlay.add_toast(Adw.Toast.new(_("Selected folder is read-only")))
            else:
                self.props.folder = response.get_path()

    @Gtk.Template.Callback()
    def remove_folder(self, *args):
        self.props.folder = None

    @Gtk.Template.Callback()
    def on_cancel(self, *args):
        self.preview_selector_button.disconnect(self._preview_update_conn)
        self.preview_selector_button.teardown()
        self.files = None
        self.set_can_close(True)
        self.close()

    @Gtk.Template.Callback()
    def do_rename(self, *args):
        has_sandboxed = self._has_sandboxed_files and not self.props.folder
        if not (self.props.validation_passed and not has_sandboxed):
            return

        self.error_banner.set_revealed(False)
        format = self.filename_entry.get_text()
        names = []
        for file in self.files:
            if self.props.folder:
                basepath = self.props.folder
            else:
                basepath = os.path.dirname(file.props.path)
            names.append(
                os.path.join(
                    basepath,
                    parse_placeholder_string(format, file) + file.props.filetype,
                )
            )

        self.set_sensitive(False)
        self.set_can_close(False)

        self._renamed = len(self.files)
        self.file_manager.rename_files(self.files, names)

    # Preview

    def generate_preview_attrs(self, file: EartagFile):
        """Generates preview attributes for a given file."""
        preview_attrs = Pango.AttrList()

        parsed_placeholder, placeholder_positions = parse_placeholder_string(
            self.filename_entry.get_text(), file, positions=True
        )
        preview_text = parsed_placeholder + file.props.filetype

        for n in range(len(placeholder_positions)):
            start, end = placeholder_positions[n]
            color = THEMES[self.syntax_highlight.props.theme]["placeholder_colors"][
                n % len(THEMES[self.syntax_highlight.props.theme]["placeholder_colors"])
            ]
            color_attr = attr_foreground_new(color, start, end)
            preview_attrs.insert(color_attr)
        return preview_text, preview_attrs

    @Gtk.Template.Callback()
    def update_preview(self, *args):
        """Validates the input and updates the preview."""
        # TODO: figure out why this is needed
        if not self.files:
            self.preview_selector_button.teardown()
            return
        self.validate_placeholder()
        self.preview_selector_button.emit("formatting-changed")
        preview_text, preview_attrs = self.generate_preview_attrs(
            self.files[self.preview_selector_button.props.selected_index]
        )
        self.preview_entry.set_text(preview_text)
        self.preview_entry.set_attributes(preview_attrs)

    def on_done(self, task, *args):
        self.set_can_close(True)

        self.parent.toast_overlay.add_toast(
            Adw.Toast.new(
                ngettext(
                    # TRANSLATORS: {renamed} is a placeholder for the number
                    # of tracks the tags were succesfully extracted for.
                    # **Do not translate the text between the curly brackets!**
                    "Renamed 1 file",
                    "Renamed {renamed} files",
                    self._renamed,
                ).format(renamed=self._renamed)
            )
        )
        self.files = None
        self.close()

    @Gtk.Template.Callback()
    def on_sandbox_warning_learn_more(self, *args):
        Gio.AppInfo.launch_default_for_uri(
            "https://gitlab.gnome.org/World/eartag/-/wikis/How-to-fix-renaming-files-not-working-under-Flatpak"  # noqa: E501
        )



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import Adw, Gtk, Gdk
from . import APP_GRESOURCE_PATH
import asyncio

from enum import IntEnum


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/dialogs/closewarning.ui")
class EartagCloseWarningDialog(Adw.AlertDialog):
    __gtype_name__ = "EartagCloseWarningDialog"

    def __init__(self, window, file_manager):
        super().__init__()
        self.window = window
        self.file_manager = file_manager

    async def handle_response_async(self, dialog, response):
        if response == "discard":
            self.window.force_close = True
            self.window.close()
        elif response == "save":
            if not await self.file_manager.save_async():
                self.close()
                return
            self.window.close()
        else:
            self.close()

    @Gtk.Template.Callback()
    def handle_response(self, dialog, response):
        asyncio.create_task(self.handle_response_async(dialog, response))


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/dialogs/discardwarning.ui")
class EartagDiscardWarningDialog(Adw.AlertDialog):
    __gtype_name__ = "EartagDiscardWarningDialog"

    def __init__(self):
        self.closed_event = asyncio.Event()
        self.response = None

    def __del__(self):
        self.closed_event.clear()
        del self.closed_event
        del self.response

    async def wait_for_response(self):
        """Wait for a response from the dialog."""
        await self.closed_event.wait()
        return self.response

    def present(self, window):
        """Show the dialog."""
        self.closed_event.clear()
        self.response = None
        super().present(window)

    @Gtk.Template.Callback()
    def handle_response(self, dialog, response):
        """Save the response from the dialog to the response property."""
        self.response = response
        self.closed_event.set()
        self.close()


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/dialogs/removaldiscardwarning.ui")
class EartagRemovalDiscardWarningDialog(Adw.AlertDialog):
    __gtype_name__ = "EartagRemovalDiscardWarningDialog"

    def __init__(self, file_manager, files):
        super().__init__()
        self.file_manager = file_manager
        self.files = files

    async def handle_response_async(self, dialog, response):
        if response == "save":
            if not await self.file_manager.save_async():
                return
        if response != "cancel":
            self.file_manager.remove_files(self.files, force_discard=True)
        self.file = None
        self.close()

    @Gtk.Template.Callback()
    def handle_response(self, dialog, response):
        asyncio.create_task(self.handle_response_async(dialog, response))


class EartagErrorType(IntEnum):
    """Error types which can be passed to the EartagErrorDialog."""

    ERROR_UNKNOWN = 0
    ERROR_LOAD = 1
    ERROR_RENAME = 2
    ERROR_SAVE = 3


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/dialogs/error.ui")
class EartagErrorDialog(Adw.AlertDialog):
    """Generic error dialog with a log view."""

    __gtype_name__ = "EartagErrorDialog"

    logs_view = Gtk.Template.Child()

    def __init__(self, error_type: EartagErrorType, logs: str):
        super().__init__()
        self.logs = logs

        if error_type == EartagErrorType.ERROR_LOAD:
            self.props.heading = _("Failed to Load Files")
            error_message = _("Some files could not be loaded.")

        elif error_type == EartagErrorType.ERROR_RENAME:
            self.props.heading = _("Failed to Rename Files")
            error_message = _("Some files could not be renamed.")

        elif error_type == EartagErrorType.ERROR_SAVE:
            self.props.heading = _("Failed to Save Files")
            error_message = _("Some files could not be saved.")

        else:
            self.props.heading = _("An Error Has Occured")
            error_message = _("An internal error has occured.")

        self.props.body = (
            error_message
            + "\n\n"
            + _(
                'Please copy the logs below and <a href="https://gitlab.gnome.org/World/eartag/-/issues/new">submit an issue report</a>.'
            )
        )

        self.logs_view.get_buffer().set_text(self.logs)

    def present(self, parent):
        self._parent = parent
        super().present(parent)

    @Gtk.Template.Callback()
    def handle_response(self, dialog, response):
        if response == "copy":
            self.get_clipboard().set_content(Gdk.ContentProvider.new_for_value(self.logs))
            self._parent.toast_overlay.add_toast(Adw.Toast.new(_("Copied error log to clipboard")))
        self.close()
        del self._parent


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/dialogs/tagdeletewarning.ui")
class EartagTagDeleteWarningDialog(Adw.AlertDialog):
    __gtype_name__ = "EartagTagDeleteWarningDialog"

    def __init__(self, window):
        super().__init__()
        self.window = window

    @Gtk.Template.Callback()
    def handle_response(self, dialog, response):
        if response != "cancel":
            self.window.do_delete_all_tags()
        self.close()



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors
"""
Common code for the preview selector in the rename and extract tags dialogs.
"""

from gi.repository import Gtk, GObject, Gio, Gdk
import os.path

from ..backends.file import EartagFile
from .. import APP_GRESOURCE_PATH


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/previewselectoritem.ui")
class EartagPreviewSelectorItem(Gtk.Box):
    """
    Representation of a file for the preview file selector in the rename
    and extract tags dialogs.
    """

    __gtype_name__ = "EartagPreviewSelectorItem"

    formatted_path_label = Gtk.Template.Child()
    filename_label = Gtk.Template.Child()

    def __init__(self, parent, formatting_function):
        """
        Initializes the EartagPreviewSelectorItem. formatting_function should be
        a function that takes an EartagFile and returns a tuple containing
        a string to display and a PangoAttrList containing its Pango attributes.
        """
        super().__init__()
        self._file = None
        self._func = formatting_function
        self._parent = parent
        self._formatting_changed_conn = self._parent.connect(
            "formatting-changed", self.update_label
        )

    @GObject.Property(type=EartagFile, default=None)
    def file(self):
        return self._file

    @file.setter
    def file(self, new_file: EartagFile):
        self._file = new_file
        if new_file is not None:
            self.update_label()

    def update_label(self, *args):
        label, attributes = self._func(self._file)
        self.formatted_path_label.props.label = label
        self.formatted_path_label.props.attributes = attributes
        self.filename_label.props.label = os.path.basename(self._file.props.path)

    def unbind(self):
        self.props.file = None
        del self._file

    def teardown(self):
        self._parent.disconnect(self._formatting_changed_conn)
        del self._func
        del self._parent


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/previewselectorbutton.ui")
class EartagPreviewSelectorButton(Gtk.MenuButton):
    """
    Previewed file selector for the rename and extract tags dialogs.
    """

    __gtype_name__ = "EartagPreviewSelectorButton"

    popover = Gtk.Template.Child()
    list = Gtk.Template.Child()
    search_entry = Gtk.Template.Child()

    selected_index = GObject.Property(type=int, default=0)

    # Emit this signal when the formatting needs to be refreshed
    formatting_changed = GObject.Signal()

    def __init__(self):
        self.search_entry.connect("search-changed", self.refresh_filter)

        self.model_nofilter = Gio.ListStore(item_type=EartagFile)
        self.model = Gtk.FilterListModel(model=self.model_nofilter)
        self.filter = Gtk.CustomFilter.new(self.filter_func, self.model)
        self.model.set_filter(self.filter)

        self.selection_model = Gtk.NoSelection.new(self.model)

        # "Add additional preview_selector" field
        # (we can reuse the factory from the "more preview_selectors" selector):
        self._ignore_preview_selector_selector = False

        self.factory = Gtk.SignalListItemFactory.new()
        self.factory.connect("setup", self.factory_setup)
        self.factory.connect("bind", self.factory_bind)
        self.factory.connect("unbind", self.factory_unbind)
        self.factory.connect("teardown", self.factory_teardown)
        self.list.set_model(self.selection_model)
        self.list.connect("activate", self.set_previewed_file_from_selection)

        # Close popover if Escape key is pressed in search entry
        controller = Gtk.ShortcutController()
        trigger = Gtk.KeyvalTrigger.new(Gdk.keyval_from_name("Escape"), 0)
        shortcut = Gtk.Shortcut.new(
            trigger, Gtk.CallbackAction.new(self.close_preview_selector_popover)
        )
        controller.add_shortcut(shortcut)
        self.search_entry.add_controller(controller)

    def set_files(self, files: list):
        self.model_nofilter.splice(0, 0, files)
        self.refresh_filter()

    def set_formatting_function(self, func):
        self._formatting_function = func
        # We do this here since otherwise it will fail as the formatting
        # function isn't set yet.
        self.list.set_factory(self.factory)

    def filter_func(self, item, *args):
        """Filter function for the tag dropdown."""
        query = self.search_entry.get_text()
        if query:
            return (
                query.lower() in f"{item.props.artist.lower()} {item.props.title.lower()}"
                or query.lower() in os.path.basename(item.path).lower()
            )
        return True

    def refresh_filter(self, *args):
        self.filter.changed(Gtk.FilterChange.DIFFERENT)

    def close_preview_selector_popover(self, *args):
        self.popover.popdown()

    def set_previewed_file_from_selection(self, list, selection, *args):
        item = self.model.get_item(selection)
        index_found, index = self.model_nofilter.find(item)
        index = index if index_found else 0
        self.props.selected_index = index
        self.close_preview_selector_popover()

    def factory_setup(self, factory, list_item):
        list_item.set_child(EartagPreviewSelectorItem(self, self._formatting_function))

    def factory_bind(self, factory, list_item):
        list_item.get_child().props.file = list_item.get_item()

    def factory_unbind(self, factory, list_item):
        list_item.get_child().unbind()
        list_item.get_child().teardown()

    def factory_teardown(self, factory, list_item):
        if list_item.get_child():
            list_item.get_child().teardown()

    def teardown(self):
        # Ensure no items are left over
        self.filter.changed(Gtk.FilterChange.DIFFERENT)
        self.filter.set_filter_func(lambda *x: False)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject
import asyncio

import mutagen.mp3
import mutagen.wave
import mutagen.id3
from mutagen.id3 import PictureType

from .file import CoverType
from .file_mutagen_common import EartagFileMutagenCommon
from ..utils.misc import safe_int
from ..utils.validation import get_mimetype, get_mimetype_buffer

# These are copied from the code for Mutagen's EasyID3 functions:
KEY_TO_FRAME = {
    "album": "TALB",
    "bpm": "TBPM",
    "compilation": "TCMP",
    "composer": "TCOM",
    "copyright": "TCOP",
    "encodedby": "TENC",
    "genre": "TCON",
    "lyricist": "TEXT",
    "length": "TLEN",
    "media": "TMED",
    "mood": "TMOO",
    "grouping": "TIT1",
    "title": "TIT2",
    "version": "TIT3",
    "artist": "TPE1",
    "albumartist": "TPE2",
    "conductor": "TPE3",
    "arranger": "TPE4",
    "discnumber": "TPOS",
    "publisher": "TPUB",
    "tracknumber": "TRCK",
    "totaltracknumber": "TRCK",
    "author": "TOLY",
    "albumartistsort": "TSO2",
    "albumsort": "TSOA",
    "composersort": "TSOC",
    "artistsort": "TSOP",
    "titlesort": "TSOT",
    "isrc": "TSRC",
    "discsubtitle": "TSST",
    "language": "TLAN",
    "comment": "COMM::eng",
    "url": "WXXX:",
}

KEY_TO_FRAME_CLASS = {
    "album": mutagen.id3.TALB,
    "bpm": mutagen.id3.TBPM,
    "compilation": mutagen.id3.TCMP,
    "composer": mutagen.id3.TCOM,
    "copyright": mutagen.id3.TCOP,
    "encodedby": mutagen.id3.TENC,
    "lyricist": mutagen.id3.TEXT,
    "length": mutagen.id3.TLEN,
    "media": mutagen.id3.TMED,
    "mood": mutagen.id3.TMOO,
    "grouping": mutagen.id3.TIT1,
    "title": mutagen.id3.TIT2,
    "version": mutagen.id3.TIT3,
    "artist": mutagen.id3.TPE1,
    "albumartist": mutagen.id3.TPE2,
    "conductor": mutagen.id3.TPE3,
    "arranger": mutagen.id3.TPE4,
    "discnumber": mutagen.id3.TPOS,
    "publisher": mutagen.id3.TPUB,
    "tracknumber": mutagen.id3.TRCK,
    "author": mutagen.id3.TOLY,
    "albumartistsort": mutagen.id3.TSO2,
    "albumsort": mutagen.id3.TSOA,
    "composersort": mutagen.id3.TSOC,
    "artistsort": mutagen.id3.TSOP,
    "titlesort": mutagen.id3.TSOT,
    "isrc": mutagen.id3.TSRC,
    "discsubtitle": mutagen.id3.TSST,
    "language": mutagen.id3.TLAN,
    "comment": mutagen.id3.COMM,
    "url": mutagen.id3.WXXX,
}

FREEFORM_KEYS = {
    "musicbrainz_artistid": "MusicBrainz Artist Id",
    "musicbrainz_albumid": "MusicBrainz Album Id",
    "musicbrainz_albumartistid": "MusicBrainz Album Artist Id",
    "musicbrainz_trackid": "MusicBrainz Release Track Id",
    "musicbrainz_releasegroupid": "MusicBrainz Release Group Id",
}


class EartagFileMutagenID3(EartagFileMutagenCommon):
    """EartagFile handler that uses mutagen for ID3 support."""

    __gtype_name__ = "EartagFileMutagenID3"
    _supports_album_covers = True
    _cover_mimetypes = ["image/jpeg", "image/png"]
    _supports_full_dates = True

    supported_extra_tags = (
        "bpm", "compilation", "composer", "copyright", "encodedby",
        "mood", "conductor", "arranger", "discnumber", "publisher",
        "isrc", "language", "discsubtitle", "url",

        "albumartistsort", "albumsort", "composersort", "artistsort",
        "titlesort",

        "musicbrainz_artistid", "musicbrainz_albumid",
        "musicbrainz_albumartistid", "musicbrainz_trackid",
        "musicbrainz_recordingid", "musicbrainz_releasegroupid"
    )  # fmt: skip

    async def load_from_file(self, path):
        mimetype = get_mimetype(path)

        self.mg_file = None
        # try:
        if mimetype in (
            "audio/mp3",
            "audio/mpeg",
            "audio/x-mpeg",
        ):
            self.mg_file = await asyncio.to_thread(mutagen.mp3.MP3, path)
        elif mimetype in ("audio/wav", "audio/x-wav"):
            self.mg_file = await asyncio.to_thread(mutagen.wave.WAVE, path)
        else:
            print(f"unknown mime {mimetype}")
        # except:
        # 	traceback.print_exc()
        # 	pass

        if self.mg_file is None:
            self.mg_file = await asyncio.to_thread(mutagen.id3.ID3, path)

        if not self.mg_file.tags:
            try:
                self.mg_file.add_tags()
            except (mutagen.id3._util.error, mutagen.wave.error):
                pass

        await self.load_cover()
        self.setup_present_extra_tags()
        self.setup_original_values()

        # Work around ID3v1 handling issues. Mutagen assumes ID3v1.1, which
        # uses the last bit of the comment field as a track number. There
        # doesn't seem to be a way to check if a file is ID3v1 or ID3v1.1,
        # but we can make an educated guess based on the length of the ID3v1
        # comment field - if it goes the entire 29 characters, then it's
        # fairly likely to be 30 characters long.
        #
        # (I assume, by the extremely vague description of the header on
        # id3.org, that an ID3v1 comment field would go up to 28 characters,
        # as the last bit would be 0 (tags are supposed to end with 0s)...)

        if (
            "COMM:ID3v1 Comment:eng" in self.mg_file.tags
            and len(self.mg_file.tags["COMM:ID3v1 Comment:eng"].text[0]) == 29
        ):
            self.set_tag("tracknumber", "")

    def get_tag(self, tag_name):
        """Gets a tag's value using the KEY_TO_FRAME list as a guideline."""
        tag_name = tag_name.lower()

        if tag_name in KEY_TO_FRAME:
            try:
                return self.mg_file.tags[KEY_TO_FRAME[tag_name]].text[0]
            except KeyError:
                return ""
        elif tag_name in FREEFORM_KEYS:
            try:
                return self.mg_file.tags["TXXX:" + FREEFORM_KEYS[tag_name]].text[0]
            except KeyError:
                return ""
        elif tag_name == "musicbrainz_recordingid":
            for key, frame in list(self.mg_file.tags.items()):
                if frame.FrameID == "UFID" and frame.owner == "http://musicbrainz.org":
                    return self.mg_file.tags[key].data.decode("ascii", "ignore")
            return ""

        raise ValueError

    def set_tag(self, tag_name, value):
        """Sets a tag's value using the KEY_TO_FRAME list as a guideline."""
        tag_name = tag_name.lower()

        if tag_name in KEY_TO_FRAME:
            frame_name = KEY_TO_FRAME[tag_name.lower()]
            frame_class = KEY_TO_FRAME_CLASS[tag_name.lower()]

            # For float values that do not have numbers after the decimal point,
            # trim the trailing .0
            if tag_name in self.float_properties and value % 1 == 0:
                stringified = str(safe_int(value))
            else:
                stringified = str(value)

            self.mg_file.tags.setall(frame_name, [frame_class(encoding=3, text=[stringified])])
        elif tag_name in FREEFORM_KEYS:
            txxx = mutagen.id3.TXXX(encoding=3, desc=FREEFORM_KEYS[tag_name], text=[value])
            self.mg_file.tags.setall("TXXX:" + FREEFORM_KEYS[tag_name], [txxx])
        elif tag_name == "musicbrainz_recordingid":
            ufid = mutagen.id3.UFID(owner="http://musicbrainz.org", data=bytes(value, "ascii"))
            self.mg_file.tags.add(ufid)
        else:
            raise ValueError

    def has_tag(self, tag_name):
        """
        Returns True or False based on whether the tag with the given name is
        present in the file.
        """
        if not self.mg_file.tags:
            return False
        if tag_name == "totaltracknumber":
            return bool(self.totaltracknumber)
        elif tag_name == "releasedate":
            return "TDRC" in self.mg_file.tags or "TDOR" in self.mg_file.tags
        elif tag_name == "url":
            return bool(self.props.url)
        elif tag_name == "comment":
            return bool(self.props.comment)
        elif tag_name in KEY_TO_FRAME:
            frame_name = KEY_TO_FRAME[tag_name.lower()]
            return frame_name in self.mg_file.tags
        elif tag_name in FREEFORM_KEYS:
            return "TXXX:" + FREEFORM_KEYS[tag_name] in self.mg_file.tags
        elif tag_name == "musicbrainz_recordingid":
            for _key, frame in list(self.mg_file.tags.items()):
                if frame.FrameID == "UFID" and frame.owner == "http://musicbrainz.org":
                    return True
            return False

        return False

    def delete_tag(self, tag_name):
        """Deletes the tag with the given name from the file."""
        if tag_name.lower() == "releasedate":
            self.mg_file.tags.delall("TDRC")
            self.mg_file.tags.delall("TDOR")
            self._releasedate_cached = ""
        elif tag_name.lower() == "url":
            self.mg_file.tags.delall("WXXX")
            self.mg_file.tags.delall("WXXX:")
            self.mg_file.tags.delall("TXXX:purl")
        elif tag_name.lower() == "comment":
            self.mg_file.tags.delall("COMM::XXX")
            self.mg_file.tags.delall("COMM::eng")
            self.mg_file.tags.delall("TXXX:comment")
        elif tag_name.lower() in KEY_TO_FRAME:
            frame_name = KEY_TO_FRAME[tag_name.lower()]
            self.mg_file.tags.delall(frame_name)
        elif tag_name.lower() in FREEFORM_KEYS:
            self.mg_file.tags.delall("TXXX:" + FREEFORM_KEYS[tag_name.lower()])
        elif tag_name.lower() == "musicbrainz_recordingid":
            for key, frame in list(self.mg_file.tags.items()):
                if frame.FrameID == "UFID" and frame.owner == "http://musicbrainz.org":
                    del self.mg_file.tags[key]

        self.mark_as_modified(tag_name, notify_prop=True)

    def delete_cover(self, cover_type: CoverType, clear_only=False):
        """Delets the cover of the specified type from the file."""
        if cover_type == CoverType.FRONT:
            pictypes = (PictureType.OTHER, PictureType.COVER_FRONT)
        elif cover_type == CoverType.BACK:
            pictypes = (PictureType.COVER_BACK,)
        else:
            raise ValueError

        for tag in dict(self.mg_file.tags).copy().keys():
            if tag.startswith("APIC") and self.mg_file.tags[tag].type in pictypes:
                del self.mg_file.tags[tag]

        if not clear_only:
            self._cleanup_cover(cover_type)

    def on_remove(self, *args):
        super().on_remove()

    async def set_cover_from_data(self, cover_type: CoverType, data: str, mime: str | None = None):
        if cover_type != CoverType.FRONT and cover_type != CoverType.BACK:
            raise ValueError

        if not mime:
            mime = get_mimetype_buffer(data)

        # Set cover in UI and check if it's valid
        ret = await self._set_cover_from_data(cover_type, data)
        if ret is False:
            return

        if cover_type == CoverType.FRONT:
            self.mg_file.tags.add(
                mutagen.id3.APIC(
                    encoding=3,
                    desc="Front Cover",
                    mime=mime,
                    data=data,
                    type=PictureType.COVER_FRONT,
                )
            )

            self.mg_file.tags.add(
                mutagen.id3.APIC(
                    encoding=3,
                    desc="Cover",
                    mime=mime,
                    data=data,
                    type=PictureType.OTHER,
                )
            )

        elif cover_type == CoverType.BACK:
            self.mg_file.tags.add(
                mutagen.id3.APIC(
                    encoding=3,
                    desc="Back Cover",
                    mime=mime,
                    data=data,
                    type=PictureType.COVER_BACK,
                )
            )

    async def load_cover(self):
        """Loads the covers from the file and saves them to a temporary file."""
        front_picture = None
        front_picture_type = None
        back_picture = None

        pictures = self.mg_file.tags.getall("APIC")

        for picture in pictures:
            if picture.type == PictureType.COVER_FRONT and (
                not front_picture or front_picture_type == PictureType.OTHER
            ):
                front_picture = picture
                front_picture_type = PictureType.COVER_FRONT

            elif picture.type == PictureType.OTHER and not front_picture:
                front_picture = picture
                front_picture_type = PictureType.OTHER

            elif picture.type == PictureType.COVER_BACK:
                back_picture = picture

        if front_picture:
            await self._set_cover_from_data(CoverType.FRONT, front_picture.data, modified=False)

        if back_picture:
            await self._set_cover_from_data(CoverType.BACK, back_picture.data, modified=False)

    @GObject.Property(type=int)
    def tracknumber(self):
        tracknum_raw = self.get_tag("tracknumber")
        if not tracknum_raw:
            return None

        if "/" in tracknum_raw:
            return safe_int(tracknum_raw.split("/")[0])
        return safe_int(tracknum_raw)

    @tracknumber.setter
    def tracknumber(self, value):
        if self.totaltracknumber:
            self.set_tag(
                "tracknumber",
                "{n}/{t}".format(n=str(value), t=str(self.totaltracknumber)),
            )
        else:
            if value:
                self.set_tag("tracknumber", value)
            elif self.has_tag("tracknumber"):
                self.delete_tag("tracknumber")
        self.mark_as_modified("tracknumber")

    @GObject.Property(type=int)
    def totaltracknumber(self):
        tracknum_raw = self.get_tag("tracknumber")
        if not tracknum_raw:
            return None

        if "/" in tracknum_raw:
            return safe_int(tracknum_raw.split("/")[1])
        return None

    @totaltracknumber.setter
    def totaltracknumber(self, value):
        if self.tracknumber:
            self.set_tag("tracknumber", "{n}/{t}".format(n=str(self.tracknumber), t=str(value)))
        else:
            if value:
                self.set_tag("tracknumber", "0/{t}".format(t=str(value)))
            elif self.has_tag("tracknumber"):
                self.delete_tag("tracknumber")
        self.mark_as_modified("totaltracknumber")

    @GObject.Property()
    def genre(self):
        if "TCON" not in self.mg_file.tags:
            return None
        genre_raw = self.mg_file.tags["TCON"]
        if genre_raw and genre_raw.genres:
            return genre_raw.genres[0]
        return None

    @genre.setter
    def genre(self, value):
        if value:
            try:
                genre_raw = self.mg_file.tags["TCON"]
            except KeyError:
                self.mg_file.tags.add(mutagen.id3.TCON(encoding=3, text=[value]))
            else:
                genre_raw.encoding = 3
                genre_raw.genres = [value]
        else:
            self.delete_tag("genre")
        self.mark_as_modified("genre")

    @GObject.Property()
    def comment(self):
        if "COMM::XXX" in self.mg_file.tags:
            return self.mg_file.tags["COMM::XXX"].text[0]
        elif "COMM::eng" in self.mg_file.tags:
            return self.mg_file.tags["COMM::eng"].text[0]
        elif "TXXX:comment" in self.mg_file.tags:
            return self.mg_file.tags["TXXX:comment"].text[0]
        return None

    @comment.setter
    def comment(self, value):
        if value:
            self.mg_file.tags.setall(
                "COMM",
                [mutagen.id3.COMM(encoding=3, lang="eng", desc="", text=[str(value)])],
            )
        else:
            self.delete_tag("comment")
        self.mark_as_modified("comment")

    # These set both TDRC (date) and TDOR (original date) for compatibility.
    @GObject.Property(type=str)
    def releasedate(self):
        if not self._releasedate_cached and self.has_tag("releasedate"):
            value = ""
            if "TDRC" in self.mg_file.tags:
                value = self.mg_file.tags["TDRC"].text[0].text
            elif "TDOR" in self.mg_file.tags:
                value = self.mg_file.tags["TDOR"].text[0].text
            self._releasedate_cached = value
        return self._releasedate_cached

    @releasedate.setter
    def releasedate(self, value):
        self.validate_date("releasedate", value)
        self._releasedate_cached = value
        if not value:
            self.delete_tag("releasedate")
        elif "releasedate" not in self._error_fields:
            self.mg_file.tags.setall("TDRC", [mutagen.id3.TDRC(encoding=3, text=[str(value)])])
            if (
                "TDOR" not in self.mg_file.tags
                or self.mg_file.tags["TDOR"] == self.mg_file.tags["TDRC"]
            ):
                self.mg_file.tags.setall("TDOR", [mutagen.id3.TDOR(encoding=3, text=[str(value)])])
        self.mark_as_modified("releasedate")

    @GObject.Property(type=int)
    def discnumber(self):
        discnum_raw = self.get_tag("discnumber")
        if not discnum_raw:
            return None

        if "/" in discnum_raw:
            return safe_int(discnum_raw.split("/")[0])
        return safe_int(discnum_raw)

    @discnumber.setter
    def discnumber(self, value):
        if value:
            self.set_tag("discnumber", f"{value}/{value}")
        elif self.has_tag("discnumber"):
            self.delete_tag("discnumber")
        self.mark_as_modified("discnumber")

    @GObject.Property(type=str)
    def url(self):
        if "WXXX" in self.mg_file.tags:
            return self.mg_file.tags["WXXX"].url
        elif "WXXX:" in self.mg_file.tags:
            return self.mg_file.tags["WXXX:"].url
        elif "TXXX:purl" in self.mg_file.tags:
            return self.mg_file.tags["TXXX:purl"].text[0]
        return None

    @url.setter
    def url(self, value):
        if value:
            self.mg_file.tags.setall(
                "TXXX:purl",
                [mutagen.id3.TXXX(encoding=3, desc="purl", text=str(value))],
            )
            self.mg_file.tags.setall(
                "WXXX", [mutagen.id3.WXXX(encoding=3, desc="", url=str(value))]
            )
            self.mark_as_modified("url")
        else:
            self.delete_tag("url")



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import Gio, GObject, GLib, Gtk
import asyncio
import aiofiles
import aiofiles.os
import mimetypes
import os.path
import stat

from .backends import (
    EartagFileMutagenVorbis,
    EartagFileMutagenID3,
    EartagFileMutagenMP4,
    EartagFileMutagenASF,
)
from .backends.file import EartagFile
from .config import config
from .utils.asynctask import EartagAsyncMultitasker
from .utils.misc import find_in_model, cleanup_filename, natural_compare, iter_selection_model
from .utils.validation import (
    get_mimetype_async,
    is_valid_music_file_async,
)
from .dialogs import EartagRemovalDiscardWarningDialog


async def eartagfile_from_path(path):
    """Returns an EartagFile subclass for the provided file."""
    if not os.path.exists(path):
        raise ValueError

    # We do two filetype guesses:
    # - One with mimetypes.guess_type; this only gets the type from the file
    #   extension;
    # - One with get_mimetype; this actually checks the file contents.
    # The magic guess is given a higher priority (2) than the mimetypes guess (1),
    # to avoid misdetections in cases where the file extension is incorrect.

    mimetypes_guess = mimetypes.guess_type(path)[0]
    magic_guess = await get_mimetype_async(path, no_extension_guess=True)

    def is_type_bulk(types):
        if magic_guess in types:
            return 2
        if mimetypes_guess in types:
            return 1
        return 0

    confidence = {
        "mp3": is_type_bulk(
            ("audio/mp3", "audio/mpeg", "audio/x-mpeg", "audio/wav", "audio/x-wav")
        ),
        "ogg": is_type_bulk(
            (
                "audio/flac",
                "audio/ogg",
                "application/ogg",
                "application/x-ogg",
                "audio/x-flac",
                "audio/x-vorbis+ogg",
            )
        ),
        "mp4": is_type_bulk(("audio/x-m4a", "audio/aac", "audio/mp4", "video/mp4")),
        "wma": is_type_bulk(("audio/x-ms-wma", "audio/wma", "video/x-ms-asf")),
    }

    filetype = sorted(confidence.keys(), key=lambda ftype: confidence[ftype], reverse=True)[0]

    if filetype == "mp3":
        return await EartagFileMutagenID3.new_from_path(path)
    elif filetype == "ogg":
        return await EartagFileMutagenVorbis.new_from_path(path)
    elif filetype == "mp4":
        return await EartagFileMutagenMP4.new_from_path(path)
    elif filetype == "wma":
        return await EartagFileMutagenASF.new_from_path(path)

    raise ValueError(
        f"Unsupported file format for file {path} (mimetype: {mimetypes_guess} / {magic_guess})"
    )  # noqa: E501


class EartagFileManager(GObject.Object):
    """Contains information about the currently loaded files."""

    # Load modes
    LOAD_OVERWRITE = 0
    LOAD_INSERT = 1

    _is_modified = False
    _has_error = False

    def __init__(self, window):
        super().__init__()
        self.window = window

        # Set up models
        self.files = Gio.ListStore(item_type=EartagFile)

        # Set up sort model for sort button
        self.file_sort_model = Gtk.SortListModel(model=self.files)
        self.sorter = Gtk.CustomSorter.new(self.file_sort_func, None)
        self.file_sort_model.set_sorter(self.sorter)

        # Set up filter model for search
        self.file_filter_model = Gtk.FilterListModel(model=self.file_sort_model)
        self.filter = Gtk.CustomFilter.new(self.file_filter_func, self.file_filter_model)
        self._file_filter_str = ""
        self.file_filter_model.set_filter(self.filter)

        # Set up selection model for selected files handling
        self.selected_files = Gtk.MultiSelection.new(self.file_filter_model)

        self.file_paths = set()
        self._files_buffer = []
        self._modified_files = []
        self._error_files = []
        self._connections = {}
        self._selected_file_ids = set()
        self._selection_removed = False
        self._is_selected_cache: dict[str, bool] = {}

        # Create Multitaskers for queued operations
        self.load_task = EartagAsyncMultitasker(self._load_single_file, workers=3)
        self.load_task.connect("task-done", self.on_file_load)
        self.load_task.connect("notify::is-running", self.notify_busy)

        self.save_task = EartagAsyncMultitasker(self._save_single_file, workers=3)
        self.save_task.connect("notify::is-running", self.notify_busy)

        self.rename_task = EartagAsyncMultitasker(self._rename_single_file, workers=3)
        self.rename_task.connect("task-done", self.refresh_state)
        self.rename_task.connect("notify::is-running", self.notify_busy)

        self.selected_files.connect("selection-changed", self._on_selection_changed)

    def update_modified_status(self, file, *args):
        """Responsible for setting the is_modified property."""
        if file.is_modified and file.id not in self._modified_files:
            self._modified_files.append(file.id)
        elif not file.is_modified and file.id in self._modified_files:
            self._modified_files.remove(file.id)

        self.set_property("is_modified", bool(self._modified_files))

    def update_error_status(self, file, *args):
        """Responsible for setting the has_error property."""
        if file.has_error and file.id not in self._error_files:
            self._error_files.append(file.id)
        elif not file.has_error and file.id in self._error_files:
            self._error_files.remove(file.id)

        self.set_property("has_error", bool(self._error_files))

    #
    # Saving
    #

    def save(self):
        """Saves changes in all files."""
        if not self.is_modified or self.has_error:
            return

        # self.save_task is set up in the init functions
        self.save_task.spawn_workers()
        self.save_task.queue_put_multiple(
            [f for f in self.files if f.is_writable and f.is_modified], mark_as_done=True
        )

    async def save_async(self) -> bool:
        """
        Async wrapper for the save function, allows for waiting for
        the saving process to end.

        Returns True if there were no errors, False otherwise.
        """
        if not self.is_modified:
            return True
        if self.has_error:
            return False

        self.save_task.spawn_workers()
        self.save_task.queue_put_multiple(
            [f for f in self.files if f.is_writable and f.is_modified], mark_as_done=True
        )
        await self.save_task.wait_for_completion_async()

        if self.save_task.errors:
            return False
        return True

    async def _save_single_file(self, file):
        """Saves changes in a single file."""
        return await asyncio.to_thread(file.save)

    #
    # Loading
    #

    max_recursion_depth_reached = GObject.Signal()

    def load_files(self, paths: list, overwrite: bool = False):
        """
        Loads files or folders from the provided paths.
        """
        asyncio.create_task(self.load_files_async(paths, overwrite=overwrite))

    async def load_files_async(self, paths: list, overwrite: bool = False):
        """
        Loads files or folders from the provided paths.
        """
        # If we're doing an open-with-overwrite, clear all files first
        if overwrite:
            self._load_mode = EartagFileManager.LOAD_OVERWRITE
            self.remove_all()
        else:
            self._load_mode = EartagFileManager.LOAD_INSERT

        # Spawn file loader process
        self.load_task.spawn_workers()

        # Fill the loading queue with paths
        dirs = set()
        for path in paths:
            _stat = await aiofiles.os.stat(path)
            if stat.S_ISREG(_stat.st_mode):
                await self.load_task.queue_put_async(path)
            elif stat.S_ISDIR(_stat.st_mode):
                dirs.add(path)

        if config["open-folders-recursively"]:
            MAX_DEPTH = 10
        else:
            MAX_DEPTH = 1

        depth = 0
        while dirs and depth < MAX_DEPTH:
            new_dirs = set()

            for path in dirs:
                for file in await aiofiles.os.listdir(path):
                    fpath = os.path.join(path, file)
                    _stat = await aiofiles.os.stat(fpath)
                    if stat.S_ISDIR(_stat.st_mode):
                        new_dirs.add(fpath)
                    elif stat.S_ISREG(_stat.st_mode) and await is_valid_music_file_async(fpath):
                        await self.load_task.queue_put_async(fpath)

            dirs = new_dirs
            depth += 1

        if MAX_DEPTH != 1 and dirs and depth >= MAX_DEPTH:
            self.emit("max-recursion-depth-reached")

        # Signify that the queue is done
        self.load_task.queue_done()

    async def _load_single_file(self, path):
        """
        Loads a single file. Used internally in _load_multiple_files, which should be
        used for all file loading operations.
        """
        if path in self.file_paths:
            return True

        _file = await eartagfile_from_path(path)

        self._connections[_file.id] = (
            _file.connect("modified", self.update_modified_status),
            _file.connect("notify::has-error", self.update_error_status),
        )

        self.files.append(_file)
        self._files_buffer.append(_file)
        self.file_paths.add(_file.path)

        return True

    def on_file_load(self, *args):
        """
        Actions done after file loading is done.

        This is called automatically after the loading task is done running.
        """
        asyncio.create_task(self.on_file_load_async())

    async def on_file_load_async(self):
        """
        Actions done after file loading is done.

        This is called automatically after the loading task is done running.
        """
        has_unwritable = False
        for file in self._files_buffer:
            if not file.is_writable:
                has_unwritable = True

        if has_unwritable:
            self.emit("has-unwritable")

        first_file = None
        if self._files_buffer:
            first_file = self._files_buffer[0]
        self._files_buffer = []

        self.refresh_state()
        if self._load_mode == self.LOAD_INSERT:
            if self.get_n_selected() < 2 and first_file:
                self.select_file(first_file, True)
        else:
            self.emit("select-first")

        del first_file

        if self._load_mode == self.LOAD_OVERWRITE:
            self.emit("refresh-needed")

    has_unwritable = GObject.Signal()

    #
    # Removal
    #

    def _remove_file(self, file):
        """
        Removes a single file from the opened file list. Used internally in remove_files;
        use that function instead.
        """

        for connection in self._connections[file.id]:
            file.disconnect(connection)
        self.file_paths.remove(file.path)

        self._removed_files_buffer.append(self.files.find(file)[1])

        pos = find_in_model(self.file_filter_model, file)
        if pos >= 0:
            self._selection_removed = True
        if file.id in self._modified_files:
            self._modified_files.remove(file.id)
        if file.id in self._error_files:
            self._error_files.remove(file.id)

        file.on_remove()

        return True

    def remove_files(self, files, force_discard=False):
        """Removes files from the opened file list."""
        if not files:
            return True

        for file in files:
            if file.is_modified and not force_discard:
                EartagRemovalDiscardWarningDialog(self, files).present(self.window)
                return False

        self._selection_removed = False
        self._removed_files_buffer = []

        # Handle remove all scenario
        if len(files) == self.files.get_n_items():
            return self.remove_all()

        # Otherwise, remove all the files separately
        for file in files:
            if not self._remove_file(file):
                return False

        # If we only have one file, then just remove it manually
        if len(files) == 1:
            self.files.remove(self._removed_files_buffer[0])
        else:
            # Split list into removed chunks, which will allow us to use .splice
            # (much faster than calling remove on each item individually)
            self._removed_files_buffer.sort()
            chunks = {}
            chunk_start = self._removed_files_buffer[0]
            prev_item = -1
            for item in self._removed_files_buffer:
                if item > prev_item + 1:
                    chunks[chunk_start] = prev_item
                    chunk_start = item
                prev_item = item
            chunks[chunk_start] = prev_item

            offset = 0
            for raw_chunk_start, chunk_end in chunks.items():
                chunk_start = raw_chunk_start - offset
                chunk_length = chunk_end - raw_chunk_start + 1
                offset += chunk_length

                self.files.splice(chunk_start, chunk_length, [])

        if self.get_n_selected() <= 0 and self.files:
            self.emit("select-first")

        self.set_property("is_modified", bool(self._modified_files))
        self.set_property("has_error", bool(self._error_files))
        self.refresh_state()

        self._selection_removed = False
        self._removed_files_buffer = []

        return True

    def remove_all(self):
        """Clear the opened file ist.."""
        for file in self.files:
            file.on_remove()

        self.files.remove_all()
        self.file_paths = set()
        self.selected_files.unselect_all()

        self._modified_files = []
        self._error_files = []
        self.set_property("is_modified", bool(self._modified_files))
        self.set_property("has_error", bool(self._error_files))

        self.refresh_state()

        return True

    #
    # Renaming
    #

    def rename_files(self, files, names):
        """Rename each file in the files list to the name in the names list."""
        self.rename_task.spawn_workers()
        if len(files) != len(names):
            raise ValueError("file and name lists are not of equal length")

        self.rename_task.queue_put_multiple(
            [(files[i], names[i]) for i in range(len(files))], mark_as_done=True
        )

    async def _rename_single_file(self, rename_info):
        """
        Rename a single file, with some harnesses to prevent potential
        data loss (for example due to overwriting an existing file).
        """
        file, new_path = rename_info

        old_path = file.props.path
        if old_path == new_path:
            return

        new_path = cleanup_filename(new_path, allow_path=True, full_path=True)

        def _new_path_fixups(new_path: str):
            if os.path.exists(new_path):
                orig_new_path = new_path
                i = 0
                while os.path.exists(new_path):
                    i += 1
                    filler = f" ({i})"
                    path_split = os.path.splitext(orig_new_path)
                    # Prevent the filename from becoming too long
                    if len(path_split[0]) > 249:
                        path_split = path_split[(-249 - len(filler)) :]
                    new_path = path_split[0] + filler + path_split[1]
            else:
                if not os.path.exists(os.path.dirname(new_path)):
                    os.makedirs(os.path.dirname(new_path), exist_ok=True)
            return new_path

        new_path = await asyncio.to_thread(_new_path_fixups, new_path)
        await file.set_path_async(new_path)

        file.notify("path")

        try:
            self.file_paths.remove(old_path)
        except KeyError:
            pass
        self.file_paths.add(new_path)

    #
    # File list sort and filter
    #

    @GObject.Property(type=str, default="")
    def file_filter_str(self):
        """String for the file filter."""
        return self._file_filter_str

    @file_filter_str.setter
    def file_filter_str(self, value: str):
        self._file_filter_str = value
        self.filter.changed(Gtk.FilterChange.DIFFERENT)

    def file_filter_func(self, file, *args):
        """Custom filter for file search."""
        query = self.props.file_filter_str
        if not query:
            return True
        query = query.casefold()

        if query in file.title.casefold():
            return True

        if query in file.artist.casefold():
            return True

        if query in file.album.casefold():
            return True

        if query in os.path.basename(file.path).casefold():
            return True

        return False

    def file_sort_func(self, a, b, *args):
        """Custom sort function implementation for file sorting."""

        # Step 1. Compare album names
        a_album = GLib.utf8_casefold(a.albumsort or a.album or "", -1)
        b_album = GLib.utf8_casefold(b.albumsort or b.album or "", -1)
        collate = natural_compare(a_album, b_album)

        # Step 2. Compare track numbers
        if collate == 0:
            if (a.tracknumber or -1) > (b.tracknumber or -1):
                collate = 1
            elif (a.tracknumber or -1) < (b.tracknumber or -1):
                collate = -1

        # Step 3. If the result is inconclusive, compare filenames
        if collate == 0:
            a_filename = GLib.utf8_casefold(os.path.basename(a.path), -1)
            b_filename = GLib.utf8_casefold(os.path.basename(b.path), -1)
            collate = natural_compare(a_filename, b_filename)

        return collate

    def get_n_files(self):
        return self.files.get_n_items()

    def select_all(self, *args):
        self.selected_files.select_all()

    def unselect_all(self, *args):
        self.selected_files.unselect_all()

    def select_file(self, file, unselect_other: bool = False):
        pos = find_in_model(self.file_filter_model, file)
        if pos >= 0:
            self.selected_files.select_item(pos, unselect_other)

    def unselect_file(self, file):
        pos = find_in_model(self.file_filter_model, file)
        if pos >= 0:
            self.selected_files.unselect_item(pos)

    def is_selected(self, file):
        return file.id in self._selected_file_ids

    def get_n_selected(self):
        return self.selected_files.get_selection().get_size()

    def _on_selection_changed(
        self, selection_model: Gtk.SelectionModel, position: int, n_items: int
    ):
        for i in range(position, position + n_items):
            file = self.selected_files.get_item(i)
            is_selected = self.selected_files.is_selected(i)

            if not is_selected and file.id in self._selected_file_ids:
                self._selected_file_ids.remove(file.id)
            elif is_selected:
                self._selected_file_ids.add(file.id)

        self.emit("selection-changed")

    def all_selected(self):
        return self.get_n_selected() == self.selected_files.get_n_items()

    @property
    def selected_files_list(self):
        return [file for file in iter_selection_model(self.selected_files)]

    @GObject.Signal
    def selection_changed(self):
        self.notify("is-selected-modified")

    @GObject.Signal
    def selection_override(self):
        """
        Internal signal used to communicate selection overrides to the sidebar.
        """
        pass

    @GObject.Signal
    def select_first(self):
        """See EartagFileList.handle_select_first"""
        self.selected_files.select_item(0, True)

    def get_first_selected(self):
        """Returns the position of the first selected item."""
        selected = self.selected_files.get_selection()
        if selected.get_size() < 1:
            return -1
        return selected.get_nth(0)

    def select_next(self, *args):
        """Selects the next item on the sidebar."""
        if self.selected_files.get_n_items() <= 1 or self.get_n_selected() > 1:
            return

        selected = self.get_first_selected()
        if selected < 0:
            return

        if selected + 1 >= self.selected_files.get_n_items():
            self.selected_files.select_item(0, True)
        else:
            self.selected_files.select_item(selected + 1, True)

    def select_previous(self, *args):
        """Selects the previous item on the sidebar."""
        if self.selected_files.get_n_items() <= 1 or self.get_n_selected() > 1:
            return

        selected = self.get_first_selected()
        if selected < 0:
            return

        if selected - 1 >= 0:
            self.selected_files.select_item(selected - 1, True)
        else:
            self.selected_files.select_item(self.selected_files.get_n_items() - 1, True)

    #
    # Miscelaneous
    #

    def close_dialog(self, dialog, *args):
        dialog.close()

    @GObject.Property(type=bool, default=False)
    def is_modified(self):
        return self._is_modified

    @is_modified.setter
    def is_modified(self, value):
        self._is_modified = value
        self.notify("is-selected-modified")

    @GObject.Property(type=bool, default=False)
    def is_selected_modified(self) -> bool:
        for file_id in self._modified_files:
            if file_id in self._selected_file_ids:
                return True
        return False

    @GObject.Property(type=bool, default=False)
    def has_error(self):
        return self._has_error

    @has_error.setter
    def has_error(self, value):
        self._has_error = value

    @GObject.Property(type=bool, default=False)
    def is_busy(self):
        return self.load_task.is_running or self.save_task.is_running or self.rename_task.is_running

    def notify_busy(self, *args):
        self.notify("is-busy")

    @GObject.Signal
    def refresh_needed(self):
        pass

    def refresh_state(self, *args):
        """Convenience function to refresh the state of the UI"""
        self.emit("refresh-needed")
